from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Optional, Sequence, Union

from typing_extensions import NotRequired, TypedDict

from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect
from great_expectations.expectations.metrics.metric_provider import MetricProvider
from great_expectations.expectations.metrics.util import MAX_RESULT_RECORDS
from great_expectations.util import get_sqlalchemy_subquery_type

if TYPE_CHECKING:
    from great_expectations.execution_engine import SqlAlchemyExecutionEngine

logger = logging.getLogger(__name__)


class MissingElementError(TypeError):
    def __init__(self):
        super().__init__(
            "The batch subquery selectable does not contain an "
            "element from which query parameters can be extracted."
        )


class MissingParameterError(ValueError):
    def __init__(self, parameter_name: str):
        super().__init__(f"Must provide `{parameter_name}` to `{self.__class__.__name__}` metric.")


class InvalidParameterTypeError(TypeError):
    def __init__(self, parameter_name: str, expected_type: type):
        super().__init__(f"`{parameter_name}` must be provided as type `{expected_type}`.")


class QueryParameters(TypedDict):
    column: NotRequired[str]
    column_A: NotRequired[str]
    column_B: NotRequired[str]
    columns: NotRequired[list[str]]


class QueryMetricProvider(MetricProvider):
    """Base class for all Query Metrics, which define metrics to construct SQL queries.

     An example of this is `query.table`,
     which takes in a SQL query & target table name, and returns the result of that query.

     In some cases, subclasses of MetricProvider, such as QueryMetricProvider, will already
     have correct values that may simply be inherited by Metric classes.

     ---Documentation---
         - https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_query_expectations

    Args:
         metric_name (str): A name identifying the metric. Metric Name must be globally unique in
             a great_expectations installation.
         domain_keys (tuple): A tuple of the keys used to determine the domain of the metric.
         value_keys (tuple): A tuple of the keys used to determine the value of the metric.
         query (str): A valid SQL query.
    """

    domain_keys = ("batch_id", "row_condition", "condition_parser")

    query_param_name: ClassVar[str] = "query"

    dialect_columns_require_subquery_aliases: ClassVar[set[GXSqlDialect]] = {
        GXSqlDialect.POSTGRESQL
    }

    @classmethod
    def _get_query_from_metric_value_kwargs(cls, metric_value_kwargs: dict) -> str:
        query_param = cls.query_param_name
        query: str | None = metric_value_kwargs.get(query_param) or cls.default_kwarg_values.get(
            query_param
        )
        if not query:
            raise MissingParameterError(query_param)
        if not isinstance(query, str):
            raise InvalidParameterTypeError(query_param, str)

        return query

    @classmethod
    def _get_parameters_dict_from_query_parameters(
        cls, query_parameters: Optional[QueryParameters]
    ) -> dict[str, Any]:
        if not query_parameters:
            return {}
        elif query_parameters and "columns" in query_parameters:
            columns = query_parameters.pop("columns")
            query_columns = {f"col_{i}": col for i, col in enumerate(columns, 1)}
            return {**query_parameters, **query_columns}
        else:
            return {**query_parameters}

    @classmethod
    def _get_substituted_batch_subquery_from_query_and_batch_selectable(
        cls,
        query: str,
        batch_selectable: sa.Selectable,
        execution_engine: SqlAlchemyExecutionEngine,
        query_parameters: Optional[QueryParameters] = None,
    ) -> str:
        parameters = cls._get_parameters_dict_from_query_parameters(query_parameters)

        if isinstance(batch_selectable, sa.Table):
            query = query.format(batch=batch_selectable, **parameters)
        elif isinstance(
            batch_selectable, (sa.sql.Select, get_sqlalchemy_subquery_type())
        ):  # specifying a row_condition returns the active batch as a Select
            # specifying an unexpected_rows_query returns the active batch as a Subquery or Alias
            # this requires compilation & aliasing when formatting the parameterized query
            batch = batch_selectable.compile(compile_kwargs={"literal_binds": True})
            # all join queries require the user to have taken care of aliasing themselves
            if "JOIN" in query.upper():
                query = query.format(batch=f"({batch})", **parameters)
            else:
                query = query.format(batch=f"({batch}) AS subselect", **parameters)
        else:
            query = query.format(batch=f"({batch_selectable})", **parameters)

        return query

    @classmethod
    def _get_sqlalchemy_records_from_substituted_batch_subquery(
        cls,
        substituted_batch_subquery: str,
        execution_engine: SqlAlchemyExecutionEngine,
    ) -> list[dict]:
        result: Union[Sequence[sa.Row[Any]], Any] = execution_engine.execute_query(
            sa.text(substituted_batch_subquery)
        ).fetchmany(MAX_RESULT_RECORDS)

        if isinstance(result, Sequence):
            return [element._asdict() for element in result]
        else:
            return [result]



import json
import pathlib

import pytest

from great_expectations.expectations.expectation_configuration import ExpectationConfiguration
from great_expectations.render.renderer.content_block.bullet_list_content_block import (
    ExpectationSuiteBulletListContentBlockRenderer,
)
from great_expectations.render.util import (
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)


@pytest.mark.unit
def test_substitute_none_for_missing():
    assert substitute_none_for_missing(kwargs={"a": 1, "b": 2}, kwarg_list=["c", "d"]) == {
        "a": 1,
        "b": 2,
        "c": None,
        "d": None,
    }

    my_kwargs = {"a": 1, "b": 2}
    assert substitute_none_for_missing(kwargs=my_kwargs, kwarg_list=["c", "d"]) == {
        "a": 1,
        "b": 2,
        "c": None,
        "d": None,
    }
    assert my_kwargs == {
        "a": 1,
        "b": 2,
    }, "substitute_none_for_missing should not change input kwargs in place."


@pytest.mark.unit
def test_parse_row_condition_string_pandas_engine():
    test_condition_string = ""
    assert parse_row_condition_string_pandas_engine(test_condition_string) == (
        "if $row_condition__0",
        {"row_condition__0": "True"},
    )

    test_condition_string = "Age in [0, 42]"
    assert parse_row_condition_string_pandas_engine(test_condition_string) == (
        "if $row_condition__0",
        {"row_condition__0": "Age in [0, 42]"},
    )

    test_condition_string = (
        "Survived == 1 and (SexCode not in (0, 7, x) | ~(Age > 50)) & not (PClass != '1st')"
    )
    assert parse_row_condition_string_pandas_engine(test_condition_string) == (
        "if $row_condition__0 and ($row_condition__1 or not ($row_condition__2)) and not ($row_condition__3)",  # noqa: E501 # FIXME CoP
        {
            "row_condition__0": "Survived == 1",
            "row_condition__1": "SexCode not in [0, 7, x]",
            "row_condition__2": "Age > 50",
            "row_condition__3": "PClass != '1st'",
        },
    )


@pytest.mark.filesystem
def test_expectations_using_expectation_definitions():
    dir_path = pathlib.Path(__file__).parent
    data_path = dir_path / "BulletListContentBlock.json"
    with open(data_path) as f:
        test_data = json.load(f)

    for expectation_dict in test_data:
        fake_expectation = ExpectationConfiguration(**expectation_dict)
        render_result = ExpectationSuiteBulletListContentBlockRenderer.render([fake_expectation])
        assert render_result is not None
        render_result = render_result.to_json_dict()
        assert isinstance(render_result, dict)
        assert "content_block_type" in render_result
        assert render_result["content_block_type"] in render_result
        assert isinstance(render_result[render_result["content_block_type"]], list)



from __future__ import annotations

import base64
import functools
import logging
import re
import urllib.parse
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Final,
    Iterable,
    Literal,
    Optional,
    Type,
    Union,
)
from urllib.parse import quote

from great_expectations._docs_decorators import deprecated_method_or_class, public_api
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import AnyUrl, errors
from great_expectations.compatibility.snowflake import (
    IS_SNOWFLAKE_INSTALLED,
    SNOWFLAKE_NOT_IMPORTED,
)
from great_expectations.compatibility.snowflake import URL as SnowflakeURL
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent import GxContextWarning, GxDatasourceWarning
from great_expectations.datasource.fluent.config_str import (
    ConfigStr,
    ConfigUri,
    _check_config_substitutions_needed,
)
from great_expectations.datasource.fluent.constants import (
    SNOWFLAKE_PARTNER_APPLICATION_CLOUD,
    SNOWFLAKE_PARTNER_APPLICATION_OSS,
)
from great_expectations.datasource.fluent.sql_datasource import (
    FluentBaseModel,
    SQLAlchemyCreateEngineError,
    SQLDatasource,
    TableAsset,
    TestConnectionError,
    to_lower_if_not_quoted,
)

if TYPE_CHECKING:
    from great_expectations.compatibility import sqlalchemy
    from great_expectations.compatibility.pydantic.networks import Parts
    from great_expectations.datasource.fluent.interfaces import (
        BatchMetadata,
    )
    from great_expectations.execution_engine import SqlAlchemyExecutionEngine

LOGGER: Final[logging.Logger] = logging.getLogger(__name__)

REQUIRED_QUERY_PARAMS: Final[Iterable[str]] = {  # errors will be thrown if any of these are missing
    "warehouse",
    "role",
}

MISSING: Final = object()  # sentinel value to indicate missing values


@functools.lru_cache(maxsize=4)
def _is_b64_encoded(sb: str | bytes) -> bool:
    """
    Check if a string or bytes is base64 encoded.
    By decoding and then encoding it, we can check if it's the same.

    Copied from: https://stackoverflow.com/a/45928164/6304433
    """
    try:
        if isinstance(sb, str):
            # If there's unicode, an exception will be thrown and the function will return false
            sb_bytes = bytes(sb, "ascii")
        elif isinstance(sb, bytes):
            sb_bytes = sb
        else:
            return False
        return base64.b64encode(base64.b64decode(sb_bytes)) == sb_bytes
    except Exception as exc:
        LOGGER.debug(f"Exception handled while checking base64 encoding - {exc!r}")
    return False


@functools.lru_cache(maxsize=4)
def _get_database_and_schema_from_path(url_path: str) -> dict[Literal["database", "schema"], str]:
    """
    Extracts the database and schema from the path of a URL.

    Raises UrlPathError if the path is missing database/schema.

    snowflake://user:password@account/database/schema
    """
    try:
        _, database, schema, *_ = url_path.split("/")
    except (ValueError, AttributeError) as e:
        LOGGER.info(f"Unable to split path - {e!r}")
        raise UrlPathError() from e
    if not database:
        raise UrlPathError(msg="missing database")
    if not schema:
        raise UrlPathError(msg="missing schema")
    return {"database": database, "schema": schema}


def _get_config_substituted_connection_string(
    datasource: SnowflakeDatasource,
    warning_msg: str = "Unable to perform config substitution",
) -> AnyUrl | None:
    if not isinstance(datasource.connection_string, ConfigUri):
        raise TypeError("Config substitution is only supported for `ConfigUri`")  # noqa: TRY003 # FIXME CoP
    if not datasource._data_context:
        warnings.warn(
            f"{warning_msg} for {datasource.connection_string.template_str}."
            " Likely missing a context.",
            category=GxContextWarning,
        )
        return None
    return datasource.connection_string.get_config_value(datasource._data_context.config_provider)


class AccountIdentifier(str):
    """
    Custom Pydantic compatible `str` type for the account identifier in
    a `SnowflakeDsn` or `ConnectionDetails`.

    https://docs.snowflake.com/en/user-guide/admin-account-identifier
    https://docs.snowflake.com/user-guide/organizations-connect

    Expected formats:
    1. Account name in your organization
        a. `<orgname>-<account_name>` - e.g. `"myOrg-my_account"`
        b. `<orgname>-<account-name>` - e.g. `"myOrg-my-account"`
    2. Account locator in a region
        a. `<account_locator>.<region>.<cloud>` - e.g. `"abc12345.us-east-1.aws"`

    The cloud group is optional if on aws but expecting it to be there makes it easier to
    distinguish formats.
    GX does not throw errors based on the regex parsing result.

    The `.` separated version of the Account name format is not supported with SQLAlchemy.
    """

    FORMATS: ClassVar[str] = "<orgname>-<account_name> or <account_locator>.<region>.<cloud>"

    FMT_1: ClassVar[str] = r"^(?P<orgname>[a-zA-Z0-9]+)[-](?P<account_name>[a-zA-Z0-9-_]+)$"
    FMT_2: ClassVar[str] = (
        r"^(?P<account_locator>[a-zA-Z0-9]+)\.(?P<region>[a-zA-Z0-9-]+)\.(?P<cloud>aws|gcp|azure)$"
    )
    PATTERN: ClassVar[re.Pattern] = re.compile(f"{FMT_1}|{FMT_2}")

    MSG_TEMPLATE: ClassVar[str] = (
        "Account identifier '{value}' does not match expected format {formats} ;"
        " it MAY be invalid. https://docs.snowflake.com/en/user-guide/admin-account-identifier"
    )

    def __init__(self, value: str) -> None:
        self._match = self.PATTERN.match(value)

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"

    @classmethod
    def __get_validators__(cls) -> Any:
        yield cls._validate

    @classmethod
    def get_schema(cls) -> dict:
        """Can be used by Pydantic models to customize the generated jsonschema."""
        return {
            "title": cls.__name__,
            "type": "string",
            "description": "Snowflake Account identifiers can be in one of two formats:"
            f" {cls.FORMATS}",
            "pattern": cls.PATTERN.pattern,
            "examples": [
                "myOrg-my_account",
                "myOrg-my-account",
                "abc12345.us-east-1.aws",
            ],
        }

    @classmethod
    def _validate(cls, value: str) -> AccountIdentifier:
        if not value:
            raise ValueError("Account identifier cannot be empty")  # noqa: TRY003 # FIXME CoP
        v = cls(value)
        if not v._match:
            LOGGER.info(
                cls.MSG_TEMPLATE.format(value=value, formats=cls.FORMATS),
            )
        return v

    @property
    def match(self) -> re.Match[str] | None:
        return self._match

    @property
    def orgname(self) -> str | None:
        """
        Part of format 1:
        * `<orgname>-<account_name>`
        * `<orgname>-<account-name>`
        """
        if self._match:
            return self._match.group("orgname")
        return None

    @property
    def account_name(self) -> str | None:
        """
        Part of format 1:
        * `<orgname>-<account_name>`
        * `<orgname>-<account-name>`
        """
        if self._match:
            return self._match.group("account_name")
        return None

    @property
    def account_locator(self) -> str | None:
        """Part of format 2: `<account_locator>.<region>.<cloud>`"""
        if self._match:
            return self._match.group("account_locator")
        return None

    @property
    def region(self) -> str | None:
        """Part of format 2: `<account_locator>.<region>.<cloud>`"""
        if self._match:
            return self._match.group("region")
        return None

    @property
    def cloud(self) -> Literal["aws", "gcp", "azure"] | None:
        """Part of format 2: `<account_locator>.<region>.<cloud>`"""
        if self._match:
            return self._match.group("cloud")  # type: ignore[return-value] # regex
        return None

    def as_tuple(
        self,
    ) -> tuple[str | None, str | None, str | None] | tuple[str | None, str | None]:
        fmt1 = (self.account_locator, self.region, self.cloud)
        if any(fmt1):
            return fmt1
        fmt2 = (self.orgname, self.account_name)
        if any(fmt2):
            return fmt2
        raise ValueError("Account identifier does not match either expected format")  # noqa: TRY003 # FIXME CoP


class _UrlPasswordError(pydantic.UrlError):
    """
    Custom Pydantic error for missing password in SnowflakeDsn.
    """

    code = "url.password"
    msg_template = "URL password invalid"


class _UrlDomainError(pydantic.UrlError):
    """
    Custom Pydantic error for missing domain in SnowflakeDsn.
    """

    code = "url.domain"
    msg_template = "URL domain invalid"


class UrlPathError(pydantic.UrlError):
    """
    Custom Pydantic error for missing path in SnowflakeDsn.
    """

    code = "url.path"
    msg_template = "URL path missing database/schema"

    def __init__(self, **ctx: Any) -> None:
        super().__init__(**ctx)


class _UrlMissingQueryError(pydantic.UrlError):
    """
    Custom Pydantic error for missing query parameters in SnowflakeDsn.
    """

    def __init__(self, **ctx: Any) -> None:
        super().__init__(**ctx)

    code = "url.query"
    msg_template = "URL query param missing"


class SnowflakeDsn(AnyUrl):
    allowed_schemes = {
        "snowflake",
    }

    @classmethod
    @override
    def validate_parts(cls, parts: Parts, validate_port: bool = True) -> Parts:
        """
        Overridden to validate additional fields outside of scheme (which is performed by AnyUrl).
        """
        user = parts["user"]
        if user is None:
            raise errors.UrlUserInfoError()

        password = parts["password"]
        if password is None:
            raise _UrlPasswordError()

        domain = parts["domain"]
        if domain is None:
            raise _UrlDomainError()

        validated_parts = AnyUrl.validate_parts(parts=parts, validate_port=validate_port)

        path: str = parts["path"]
        # raises UrlPathError if path is missing database/schema
        _get_database_and_schema_from_path(path)

        return validated_parts

    @property
    def params(self) -> dict[str, list[str]]:
        """The query parameters as a dictionary."""
        if not self.query:
            return {}
        return urllib.parse.parse_qs(self.query)

    @property
    def account_identifier(self) -> AccountIdentifier:
        """Alias for host."""
        assert self.host
        return AccountIdentifier(self.host)

    @property
    def account(self) -> AccountIdentifier:
        """Alias for account_identifier."""
        return self.account_identifier

    @property
    def database(self) -> str:
        assert self.path
        return self.path.split("/")[1]

    @property
    def schema_(self) -> str:
        assert self.path
        return self.path.split("/")[2]

    @property
    def warehouse(self) -> str | None:
        return self.params.get("warehouse", [None])[0]

    @property
    def role(self) -> str | None:
        return self.params.get("role", [None])[0]


class ConnectionDetails(FluentBaseModel):
    """
    Information needed to connect to a Snowflake database.
    Alternative to a connection string.

    https://docs.snowflake.com/en/developer-guide/python-connector/sqlalchemy#additional-connection-parameters
    """

    account: AccountIdentifier
    user: str
    password: Union[ConfigStr, str]
    database: str = pydantic.Field(
        ...,
        description="`database` that the Datasource is mapped to.",
    )
    schema_: str = pydantic.Field(
        ..., alias="schema", description="`schema` that the Datasource is mapped to."
    )  # schema is a reserved attr in BaseModel
    warehouse: str
    role: str
    numpy: bool = False

    @classmethod
    def required_fields(cls) -> list[str]:
        """Returns the required fields for this model as defined in the schema."""
        return cls.schema()["required"]

    class Config:
        @staticmethod
        def schema_extra(schema: dict, model: type[ConnectionDetails]) -> None:
            schema["properties"]["account"] = AccountIdentifier.get_schema()


@public_api
class SnowflakeDatasource(SQLDatasource):
    """Adds a Snowflake datasource to the data context.

    Args:
        name: The name of this Snowflake datasource.
        connection_string: The SQLAlchemy connection string used to connect to the Snowflake database.
            For example: "snowflake://<user_login_name>:<password>@<account_identifier>"
        assets: An optional dictionary whose keys are TableAsset or QueryAsset names and whose values
            are TableAsset or QueryAsset objects.
    """  # noqa: E501 # FIXME CoP

    type: Literal["snowflake"] = "snowflake"  # type: ignore[assignment] # FIXME CoP
    # TODO: rename this to `connection` for v1?
    connection_string: Union[ConnectionDetails, ConfigUri, SnowflakeDsn]  # type: ignore[assignment] # Deviation from parent class as individual args are supported for connection

    # TODO: add props for user, password, etc?

    @property
    def account(self) -> AccountIdentifier | None:
        """Convenience property to get the `account` regardless of the connection string format."""
        if isinstance(self.connection_string, (ConnectionDetails, SnowflakeDsn)):
            return self.connection_string.account

        subbed_str: str | None = _get_config_substituted_connection_string(
            self, warning_msg="Unable to determine account"
        )
        if not subbed_str:
            return None
        hostname = urllib.parse.urlparse(subbed_str).hostname
        if hostname:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=GxDatasourceWarning)
            return AccountIdentifier(hostname)
        return None

    @property
    def schema_(self) -> str | None:
        """
        Convenience property to get the `schema` regardless of the connection string format.

        `schema_` to avoid conflict with Pydantic models schema property.
        """
        if isinstance(self.connection_string, (ConnectionDetails, SnowflakeDsn)):
            return to_lower_if_not_quoted(self.connection_string.schema_)

        subbed_str: str | None = _get_config_substituted_connection_string(
            self, warning_msg="Unable to determine schema"
        )
        if not subbed_str:
            return None
        url_path: str = urllib.parse.urlparse(subbed_str).path

        return to_lower_if_not_quoted(_get_database_and_schema_from_path(url_path)["schema"])

    @property
    def database(self) -> str | None:
        """Convenience property to get the `database` regardless of the connection string format."""
        if isinstance(self.connection_string, (ConnectionDetails, SnowflakeDsn)):
            return self.connection_string.database

        subbed_str: str | None = _get_config_substituted_connection_string(
            self, warning_msg="Unable to determine database"
        )
        if not subbed_str:
            return None
        url_path: str = urllib.parse.urlparse(subbed_str).path
        return _get_database_and_schema_from_path(url_path)["database"]

    @property
    def warehouse(self) -> str | None:
        """
        Convenience property to get the `warehouse` regardless of the connection string format.
        """
        if isinstance(self.connection_string, (ConnectionDetails, SnowflakeDsn)):
            return self.connection_string.warehouse

        subbed_str: str | None = _get_config_substituted_connection_string(
            self, warning_msg="Unable to determine warehouse"
        )
        if not subbed_str:
            return None
        return urllib.parse.parse_qs(urllib.parse.urlparse(subbed_str).query).get(
            "warehouse", [None]
        )[0]

    @property
    def role(self) -> str | None:
        """Convenience property to get the `role` regardless of the connection string format."""
        if isinstance(self.connection_string, (ConnectionDetails, SnowflakeDsn)):
            return self.connection_string.role

        subbed_str: str | None = _get_config_substituted_connection_string(
            self, warning_msg="Unable to determine role"
        )
        if not subbed_str:
            return None
        return urllib.parse.parse_qs(urllib.parse.urlparse(subbed_str).query).get("role", [None])[0]

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the SnowflakeDatasource.

        Args:
            test_assets: If assets have been passed to the SQLDatasource,
            whether to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        try:
            super().test_connection(test_assets=test_assets)
        except TestConnectionError as e:
            if self.account and not self.account.match:
                raise TestConnectionError(
                    message=e.__class__.__name__,
                    addendum=AccountIdentifier.MSG_TEMPLATE.format(
                        value=self.account,
                        formats=AccountIdentifier.FORMATS,
                    ),
                ) from e
            raise

    @deprecated_method_or_class(
        version="1.0.0a4",
        message="`schema_name` is deprecated. The schema now comes from the datasource.",
    )
    @public_api
    @override
    def add_table_asset(
        self,
        name: str,
        table_name: str = "",
        schema_name: Optional[str] = MISSING,  # type: ignore[assignment] # sentinel value
        batch_metadata: Optional[BatchMetadata] = None,
    ) -> TableAsset:
        """Adds a table asset to this datasource.

        Args:
            name: The name of this table asset.
            table_name: The table where the data resides.
            schema_name: The schema that holds the table. Will use the datasource schema if not
                provided.
            batch_metadata: BatchMetadata we want to associate with this DataAsset and all batches
                derived from it.

        Returns:
            The table asset that is added to the datasource.
            The type of this object will match the necessary type for this datasource.
        """
        if schema_name is MISSING:
            # using MISSING to indicate that the user did not provide a value
            schema_name = self.schema_
        else:
            # deprecated-v0.18.16
            warnings.warn(
                "The `schema_name argument` is deprecated and will be removed in a future release."
                " The schema now comes from the datasource.",
                category=DeprecationWarning,
            )
            if schema_name != self.schema_:
                warnings.warn(
                    f"schema_name {schema_name} does not match datasource schema {self.schema_}",
                    category=GxDatasourceWarning,
                )

        return super().add_table_asset(
            name=name,
            table_name=table_name,
            schema_name=schema_name,
            batch_metadata=batch_metadata,
        )

    @pydantic.root_validator(pre=True)
    def _convert_root_connection_detail_fields(cls, values: dict) -> dict:
        """
        Convert root level connection detail fields to a ConnectionDetails compatible object.
        This preserves backwards compatibility with the previous implementation of SnowflakeDatasource.

        It also allows for users to continue to provide connection details in the
        `context.data_sources.add_snowflake()` factory functions without nesting it in a
        `connection_string` dict.
        """  # noqa: E501 # FIXME CoP
        connection_detail_fields: set[str] = {
            "schema",  # field name in ConnectionDetails is schema_ (with underscore)
            *ConnectionDetails.__fields__.keys(),
        }

        connection_string: Any | None = values.get("connection_string")
        provided_fields = tuple(values.keys())

        connection_details = {}
        for field_name in provided_fields:
            if field_name in connection_detail_fields:
                if connection_string:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        "Provided both connection detail keyword args and `connection_string`."
                    )
                connection_details[field_name] = values.pop(field_name)
        if connection_details:
            values["connection_string"] = connection_details
        return values

    @pydantic.validator("connection_string", pre=True)
    def _check_config_template(cls, connection_string: Any) -> Any:
        """
        If connection_string has a config template, parse it as a ConfigUri, ignore other errors.
        """
        if isinstance(connection_string, str):
            if ConfigUri.str_contains_config_template(connection_string):
                LOGGER.debug("`connection_string` contains config template")
                return pydantic.parse_obj_as(ConfigUri, connection_string)
        return connection_string

    @pydantic.root_validator
    def _check_xor_input_args(cls, values: dict) -> dict:
        # keeping this validator isn't strictly necessary, but it provides a better error message
        connection_string: str | ConfigUri | ConnectionDetails | None = values.get(
            "connection_string"
        )
        if connection_string:
            # Method 1 - connection string
            is_connection_string: bool = isinstance(
                connection_string, (str, ConfigStr, SnowflakeDsn)
            )
            # Method 2 - individual args (account, user, and password are bare minimum)
            has_min_connection_detail_values: bool = isinstance(
                connection_string, ConnectionDetails
            ) and bool(
                connection_string.account and connection_string.user and connection_string.password
            )
            if is_connection_string or has_min_connection_detail_values:
                return values
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            "Must provide either a connection string or"
            f" a combination of {', '.join(ConnectionDetails.required_fields())} as keyword args."
        )

    @pydantic.validator("connection_string")
    def _check_for_required_query_params(
        cls, connection_string: ConnectionDetails | SnowflakeDsn | ConfigUri
    ) -> ConnectionDetails | SnowflakeDsn | ConfigUri:
        """
        If connection_string is a SnowflakeDsn,
        check for required query parameters according to `REQUIRED_QUERY_PARAMS`.
        """
        if not isinstance(connection_string, (SnowflakeDsn, ConfigUri)):
            return connection_string

        missing_keys: set[str] = set(REQUIRED_QUERY_PARAMS)

        if connection_string.query:
            query_params: dict[str, list[str]] = urllib.parse.parse_qs(connection_string.query)

            for key in REQUIRED_QUERY_PARAMS:
                if key in query_params:
                    missing_keys.remove(key)

        if missing_keys:
            raise _UrlMissingQueryError(
                msg=f"missing {', '.join(sorted(missing_keys))}",
            )
        return connection_string

    @pydantic.validator("kwargs")
    def _base64_encode_private_key(cls, kwargs: dict) -> dict:
        if connect_args := kwargs.get("connect_args", {}):
            if private_key := connect_args.get("private_key"):
                # test if it's already base64 encoded
                if _is_b64_encoded(private_key):
                    LOGGER.info("private_key is already base64 encoded")
                else:
                    LOGGER.info("private_key is not base64 encoded, encoding now")
                    connect_args["private_key"] = base64.standard_b64encode(private_key)
        return kwargs

    class Config:
        @staticmethod
        def schema_extra(schema: dict, model: type[SnowflakeDatasource]) -> None:
            """
            Customize jsonschema for SnowflakeDatasource.
            https://docs.pydantic.dev/1.10/usage/schema/#schema-customization
            Change connection_string to be a string or a dict, but not both.
            """
            connection_string_prop = schema["properties"]["connection_string"].pop("anyOf")
            schema["properties"]["connection_string"].update({"oneOf": connection_string_prop})

    def _get_connect_args(self) -> dict[str, str | bool]:
        excluded_fields: set[str] = set(SQLDatasource.__fields__.keys())
        # dump as json dict to force serialization of things like AnyUrl
        return self._json_dict(exclude=excluded_fields, exclude_none=True)

    def _get_snowflake_partner_application(self) -> str:
        """
        This is used to set the application query parameter in the Snowflake connection URL,
        which provides attribution to GX for the Snowflake Partner program.
        """

        # This import is here to avoid a circular import
        from great_expectations.data_context import CloudDataContext

        if isinstance(self._data_context, CloudDataContext):
            return SNOWFLAKE_PARTNER_APPLICATION_CLOUD
        return SNOWFLAKE_PARTNER_APPLICATION_OSS

    def _get_url_args(self) -> dict[str, str | bool]:
        excluded_fields: set[str] = set(SQLDatasource.__fields__.keys())
        # dump as json dict to force serialization of things like AnyUrl
        return self._json_dict(exclude=excluded_fields, exclude_none=True)

    @override
    def get_execution_engine(self) -> SqlAlchemyExecutionEngine:
        """
        Overrides get_execution_engine in Datasource
        Standard behavior is to assume all top-level Datasource config (unless part of `cls._EXTRA_EXCLUDED_EXEC_ENG_ARGS`)
        should be passed to the GX ExecutionEngine constructor.

        for SQLAlchemy this would lead to creating 2 different `sqlalchemy.engine.Engine` objects
        one for the Datasource and one for the ExecutionEngine. This is wasteful and causes multiple connections to
        the database to be created.

        For Snowflake specifically we may represent the connection_string as a dict, which is not supported by SQLAlchemy.
        """  # noqa: E501 # FIXME CoP
        gx_execution_engine_type: Type[SqlAlchemyExecutionEngine] = self.execution_engine_type

        connection_string: str | None = (
            self.connection_string if isinstance(self.connection_string, str) else None
        )

        gx_exec_engine = gx_execution_engine_type(
            self.name,
            connection_string=connection_string,
            engine=self.get_engine(),
            create_temp_table=self.create_temp_table,
            data_context=self._data_context,
        )
        self._execution_engine = gx_exec_engine
        return gx_exec_engine

    @override
    def get_engine(self) -> sqlalchemy.Engine:
        if self.connection_string != self._cached_connection_string or not self._engine:
            try:
                model_dict = self.dict(
                    exclude=self._get_exec_engine_excludes(),
                    config_provider=self._config_provider,
                )
                _check_config_substitutions_needed(
                    self, model_dict, raise_warning_if_provider_not_present=True
                )

                kwargs = model_dict.pop("kwargs", {})
                connection_string: str | dict = model_dict.pop("connection_string")

                if isinstance(connection_string, str):
                    url = sa.engine.url.make_url(connection_string)
                    url = url.update_query_dict(
                        query_parameters={"application": self._get_snowflake_partner_application()}
                    )
                    self._engine = self._build_engine_with_connect_args(url=url, **kwargs)
                else:
                    self._engine = self._build_engine_with_connect_args(
                        application=self._get_snowflake_partner_application(),
                        **connection_string,
                        **kwargs,
                    )

            except Exception as e:
                # connection_string passed pydantic validation, but fails to create sqla engine
                # Possible case is a missing plugin (e.g. snowflake-sqlalchemy)
                if IS_SNOWFLAKE_INSTALLED:
                    raise SQLAlchemyCreateEngineError(
                        cause=e, addendum=str(SNOWFLAKE_NOT_IMPORTED)
                    ) from e
                raise SQLAlchemyCreateEngineError(cause=e) from e
            # connection string isn't strictly required for Snowflake, so we conditionally cache
            if isinstance(self.connection_string, str):
                self._cached_connection_string = self.connection_string
        return self._engine

    def _build_engine_with_connect_args(
        self,
        url: SnowflakeURL | None = None,
        connect_args: dict[str, Any] | None = None,
        **kwargs,
    ) -> sqlalchemy.Engine:
        if not url:
            url_args = self._get_url_args()
            url_args.update(kwargs)

            password = url_args.get("password")
            if isinstance(password, str):
                url_args["password"] = quote(password)

            url = SnowflakeURL(**url_args)
        else:
            url_args = {}

        engine_kwargs: dict[Literal["url", "connect_args"], Any] = {}
        if connect_args:
            if private_key := connect_args.get("private_key"):
                url_args.pop(  # TODO: update models + validation to handle this
                    "password", None
                )
                LOGGER.info(
                    "private_key detected,"
                    " ignoring password and using private_key for authentication"
                )
                # assume the private_key is base64 encoded
                connect_args["private_key"] = base64.standard_b64decode(private_key)

            engine_kwargs["connect_args"] = connect_args

        engine_kwargs["url"] = url

        return sa.create_engine(**engine_kwargs)  # type: ignore[misc] # FIXME CoP



from __future__ import annotations

import json
import pathlib
import sys
from pprint import pformat as pf
from typing import Any, Generator, Type

import pandas
import pytest
from packaging.version import Version

from great_expectations.datasource.fluent import (
    _PANDAS_SCHEMA_VERSION,  # this is the version we run in the standard test pipeline. Update as needed  # noqa: E501 # FIXME CoP
    _SCHEMAS_DIR,
    DataAsset,
    Datasource,
)
from great_expectations.datasource.fluent.sources import (
    _iter_all_registered_types,
)

PANDAS_VERSION: str = pandas.__version__
PYTHON_VERSION: Version = Version(f"{sys.version_info.major}.{sys.version_info.minor}")


def min_supported_python() -> Version:
    return Version("3.9")


def _models_and_schema_dirs() -> Generator[
    tuple[Type[Datasource | DataAsset], pathlib.Path, str], None, None
]:
    datasource: Type[Datasource] = Datasource
    ds_type_name: str = ""

    yield Datasource, _SCHEMAS_DIR, Datasource.__name__

    for name, model in _iter_all_registered_types():
        if issubclass(model, Datasource):
            datasource = model
            ds_type_name = name
            schema_dir = _SCHEMAS_DIR
        else:
            schema_dir = _SCHEMAS_DIR / datasource.__name__

        yield model, schema_dir, f"{ds_type_name}:{model.__name__}"


@pytest.mark.skipif(
    min_supported_python() < PYTHON_VERSION,
    reason=f"_sort_any_of() keys needs to be fixed for py {PYTHON_VERSION}",
)
@pytest.mark.timeout(
    2.0  # this is marked as unit so that it will run on different versions of python
)
@pytest.mark.unit
@pytest.mark.parametrize(
    ["fluent_ds_or_asset_model", "schema_dir"],
    [pytest.param(t[0], t[1], id=t[2]) for t in _models_and_schema_dirs()],
)
def test_vcs_schemas_match(  # noqa: C901 # FIXME CoP
    fluent_ds_or_asset_model: Type[Datasource | DataAsset], schema_dir: pathlib.Path
):
    """
    Test that json schemas for each DataSource and DataAsset match the current schema
    under version control.

    This is important because some of these classes are generated dynamically and
    monitoring these schemas files can clue us into problems that could otherwise go
    unnoticed.

    If this test is failing run `invoke schema --sync` to update schemas and commit the
    changes.

    Note: if the installed version of pandas doesn't match the one used in the standard
    test pipeline, the test will be marked a `xfail` because the schemas will not match.
    """

    def _sort_any_of(d: dict) -> str:
        if items := d.get("items"):
            _sort_any_of(items)
        if ref := d.get("$ref"):
            return ref
        if type_ := d.get("type"):
            return type_
        # return any string for sorting
        return "z"

    def _sort_lists(schema_as_dict: dict) -> None:
        """Sometimes "required" and "anyOf" lists come unsorted, causing misleading assertion failures; this corrects the issue.

        Args:
            schema_as_dict: source dictionary (will be modified "in-situ")

        """  # noqa: E501 # FIXME CoP
        key: str
        value: Any

        for key, value in schema_as_dict.items():
            if key == "required":
                schema_as_dict[key] = sorted(value)
            elif key == "anyOf":
                if isinstance(value, list):
                    for val in value:
                        if isinstance(value, dict):
                            schema_as_dict[key] = sorted(val, key=_sort_any_of)
                else:
                    schema_as_dict[key] = sorted(value, key=_sort_any_of)

            if isinstance(value, dict):
                _sort_lists(schema_as_dict=value)

    if "Pandas" in str(schema_dir) and _PANDAS_SCHEMA_VERSION != PANDAS_VERSION:
        pytest.xfail(reason=f"schema generated with pandas {_PANDAS_SCHEMA_VERSION}")

    print(f"python version: {sys.version.split()[0]}")
    print(f"pandas version: {PANDAS_VERSION}")

    schema_path = schema_dir.joinpath(f"{fluent_ds_or_asset_model.__name__}.json")
    print(schema_path)

    json_str = schema_path.read_text().rstrip()

    schema_as_dict = json.loads(json_str)
    _sort_lists(schema_as_dict=schema_as_dict)
    # we have tuples in our schema, which are mutated to lists when dumped to json
    # dump and reload the schema dict to ensure we are comparing
    fluent_ds_or_asset_model_as_dict = json.loads(fluent_ds_or_asset_model.schema_json())
    _sort_lists(schema_as_dict=fluent_ds_or_asset_model_as_dict)

    if "Excel" in str(schema_path):
        pytest.xfail(reason="Sorting of nested anyOf key")

    assert schema_as_dict == fluent_ds_or_asset_model_as_dict, (
        "Schemas are out of sync. Run `invoke schema --sync`. Also check your pandas version."
    )


@pytest.mark.big
@pytest.mark.skipif(
    _PANDAS_SCHEMA_VERSION != PANDAS_VERSION,
    reason=f"schemas generated with pandas {_PANDAS_SCHEMA_VERSION}",
)
def test_no_orphaned_schemas():
    """Ensure that there are no schemas that have no corresponding type."""
    print(f"python version: {sys.version.split()[0]}")
    print(f"pandas version: {PANDAS_VERSION}\n")

    # NOTE: this is a very low fidelity check
    all_schemas: set[str] = {t[1].__name__ for t in _iter_all_registered_types()}
    all_schemas.add("BatchRequest")
    all_schemas.add(Datasource.__name__)

    orphans: list[pathlib.Path] = []

    for schema in _SCHEMAS_DIR.glob("**/*.json"):
        if schema.stem not in all_schemas:
            orphans.append(schema)

    assert not orphans, (
        f"The following schemas appear to be orphaned and should be removed. "
        f"Run `invoke schema --sync --clean`\n{pf(orphans)}"
    )


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



from __future__ import annotations

import inspect
from inspect import Parameter, Signature
from typing import Callable, Type


def _merge_signatures(
    target: Callable,
    source: Callable,
    exclude: set[str] | None = None,
    return_type: Type | None = None,
) -> Signature:
    """
    Merge 2 method or function signatures.

    The returned signature will retain all positional and keyword arguments from
    `target` as well the return type.

    The `source` signature cannot contain any positional arguments, and will be
    'appended' to the `target` signatures arguments.

    Note: Signatures are immutable, a new signature is returned by this function.
    """
    target_sig: Signature = inspect.signature(target)
    source_sig: Signature = inspect.signature(source)

    final_exclude: set[str] = set()

    final_params: list[Parameter] = []
    for name, param in target_sig.parameters.items():
        if param.kind == Parameter.VAR_KEYWORD:
            continue
        final_exclude.add(name)
        final_params.append(param)

    if exclude:
        final_exclude.update(exclude)

    final_params.extend(
        [
            p
            for (name, p) in source_sig.parameters.items()
            if p.kind != Parameter.VAR_POSITIONAL and name not in final_exclude
        ]
    )

    return Signature(
        parameters=final_params,
        return_annotation=return_type if return_type else target_sig.return_annotation,
    )



"""Tests to ensure we are reusing BatchTestSetup instances across tests when appropriate.

The tests here work together to give some confidence that we aren't running the same setup/teardown
multiple times for equivalent TestConfigs. This is primarily to prevent regressions in this process.

Note that we aren't testing to ensure we don't over-reuse BatchTestSetup instances, e.g. for
different TestConfigs; that would be caught by our regular tests.
"""

from dataclasses import dataclass
from functools import cache
from typing import Mapping, Optional

import pandas as pd
import pytest
import sqlalchemy.dialects.postgresql as POSTGRESQL_TYPES

from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.datasource.fluent.pandas_datasource import DataFrameAsset
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)


@dataclass
class SetupTeardownCounts:
    setup_count = 0
    teardown_count = 0


class DummyTestConfig(DataSourceTestConfig):
    @property
    @override
    def label(self) -> str:
        return "test"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.unit

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        return DummyBatchTestSetup(data=data, config=self, context=context)


class DummyBatchTestSetup(BatchTestSetup[DummyTestConfig, DataFrameAsset]):
    @override
    def make_asset(self) -> DataFrameAsset:
        return self.context.data_sources.add_pandas(
            self._random_resource_name()
        ).add_dataframe_asset(self._random_resource_name())

    @override
    def make_batch(self) -> Batch:
        return (
            self.make_asset()
            .add_batch_definition_whole_dataframe(self._random_resource_name())
            .get_batch(batch_parameters={"dataframe": self.data})
        )

    @override
    def setup(self) -> None:
        counts = DummyBatchTestSetup._get_setup_teardown_counts()
        if counts.setup_count:
            assert False, "Setup is not being cached"
        counts.setup_count += 1

    @override
    def teardown(self) -> None:
        counts = DummyBatchTestSetup._get_setup_teardown_counts()
        if counts.teardown_count:
            assert False, "Teardown is not being cached"
        counts.teardown_count -= 1

    @staticmethod
    @cache
    def _get_setup_teardown_counts() -> SetupTeardownCounts:
        return SetupTeardownCounts()


@parameterize_batch_for_data_sources(
    data_source_configs=[
        DummyTestConfig(column_types={"a": POSTGRESQL_TYPES.INTEGER}),
        DummyTestConfig(column_types={"a": POSTGRESQL_TYPES.INTEGER}),
    ],
    data=pd.DataFrame({"a": [1, 2]}),
)
def test_caching_within_a_test(batch_for_datasource) -> None:
    # This should fail in setup or teardown if the setup and teardown are not being cached
    ...


@parameterize_batch_for_data_sources(
    data_source_configs=[
        DummyTestConfig(column_types={"a": POSTGRESQL_TYPES.INTEGER}),
    ],
    data=pd.DataFrame({"a": [1, 2]}),
)
def test_caching_across_tests(batch_for_datasource) -> None:
    # This should fail in setup or teardown if the setup and teardown are not being cached
    ...



import sys
from typing import List, Union

import pytest

from great_expectations.core.run_identifier import RunIdentifier
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.render.util import (
    _convert_unexpected_indices_to_df,
    build_count_and_index_table,
    build_count_table,
    num_to_str,
    resource_key_passes_run_name_filter,
    truncate_list_of_indices,
)

# module level markers
pytestmark = pytest.mark.unit


def test_num_to_str():
    f = 0.99999999999999
    # We can round
    assert num_to_str(f, precision=4) == "≈1"
    # Specifying extra precision should not cause a problem
    assert num_to_str(f, precision=20) == "0.99999999999999"

    f = 1234567890.123456  # Our float can only hold 17 significant digits
    assert num_to_str(f, precision=4) == "≈1.235e+9"
    assert num_to_str(f, precision=20) == "1234567890.123456"
    assert num_to_str(f, use_locale=True, precision=40) == "1,234,567,890.123456"

    f = 1.123456789012345  # 17 sig digits mostly after decimal
    assert num_to_str(f, precision=5) == "≈1.1235"
    assert num_to_str(f, precision=20) == "1.123456789012345"

    f = 0.1  # A classic difficulty for floating point arithmetic
    assert num_to_str(f) == "0.1"
    assert num_to_str(f, precision=20) == "0.1"
    assert num_to_str(f, no_scientific=True) == "0.1"

    f = 1.23456789012345e-10  # significant digits can come late
    assert num_to_str(f, precision=20) == "1.23456789012345e-10"
    assert num_to_str(f, precision=5) == "≈1.2346e-10"
    assert num_to_str(f, precision=20, no_scientific=True) == "0.000000000123456789012345"
    assert num_to_str(f, precision=5, no_scientific=True) == "≈0.00000000012346"

    f = 100.0  # floats should have trailing digits and numbers stripped
    assert num_to_str(f, precision=10, no_scientific=True) == "100"
    assert num_to_str(f, precision=10) == "100"
    assert num_to_str(f, precision=10, use_locale=True) == "100"

    f = 100  # integers should never be stripped!
    assert num_to_str(f, precision=10, no_scientific=True) == "100"
    assert num_to_str(f, precision=10) == "100"
    assert num_to_str(f, precision=10, use_locale=True) == "100"

    f = 1000  # If we have a number longer than our precision, we should still be able to correctly format  # noqa: E501 # FIXME CoP
    assert num_to_str(f, precision=4) == "1000"
    assert num_to_str(f) == "1000"


def test_resource_key_passes_run_name_filter():
    resource_key = ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier("test_suite"),
        run_id=RunIdentifier(run_name="foofooprofilingfoo"),
        batch_identifier="f14c3d2f6e8028c2db0c25edabdb0d61",
    )

    assert (
        resource_key_passes_run_name_filter(resource_key, run_name_filter={"equals": "profiling"})
        is False
    )
    assert (
        resource_key_passes_run_name_filter(
            resource_key, run_name_filter={"equals": "foofooprofilingfoo"}
        )
        is True
    )

    assert (
        resource_key_passes_run_name_filter(
            resource_key, run_name_filter={"not_equals": "profiling"}
        )
        is True
    )
    assert (
        resource_key_passes_run_name_filter(
            resource_key, run_name_filter={"not_equals": "foofooprofilingfoo"}
        )
        is False
    )

    assert (
        resource_key_passes_run_name_filter(resource_key, run_name_filter={"includes": "profiling"})
        is True
    )
    assert (
        resource_key_passes_run_name_filter(resource_key, run_name_filter={"includes": "foobar"})
        is False
    )

    assert (
        resource_key_passes_run_name_filter(
            resource_key, run_name_filter={"not_includes": "foobar"}
        )
        is True
    )
    assert (
        resource_key_passes_run_name_filter(
            resource_key, run_name_filter={"not_includes": "profiling"}
        )
        is False
    )

    assert (
        resource_key_passes_run_name_filter(
            resource_key,
            run_name_filter={"matches_regex": "(foo){2}profiling(foo)+"},
        )
        is True
    )
    assert (
        resource_key_passes_run_name_filter(
            resource_key,
            run_name_filter={"matches_regex": "(foo){3}profiling(foo)+"},
        )
        is False
    )


def test_convert_unexpected_indices_to_df():
    unexpected_index_list = [
        {"animals": "giraffe", "pk_1": 3, "pk_2": "three"},
        {"animals": "lion", "pk_1": 4, "pk_2": "four"},
        {"animals": "zebra", "pk_1": 5, "pk_2": "five"},
    ]
    unexpected_index_column_names = ["pk_1", "pk_2"]

    partial_unexpected_counts = [
        {"count": 1, "value": "giraffe"},
        {"count": 1, "value": "lion"},
        {"count": 1, "value": "zebra"},
    ]

    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=unexpected_index_list,
        unexpected_index_column_names=unexpected_index_column_names,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["pk_1", "pk_2", "Count"]
    assert res.iloc[0].tolist() == ["3", "three", 1]
    assert res.iloc[1].tolist() == ["4", "four", 1]
    assert res.iloc[2].tolist() == ["5", "five", 1]


def test_convert_unexpected_indices_to_df_multiple():
    result = [
        {"animals": "giraffe", "pk_1": 3},
        {"animals": "lion", "pk_1": 4},
        {"animals": "zebra", "pk_1": 5},
        {"animals": "zebra", "pk_1": 6},
        {"animals": "zebra", "pk_1": 7},
        {"animals": "zebra", "pk_1": 8},
    ]
    unexpected_index_column_names = ["pk_1"]
    partial_unexpected_counts = [
        {"count": 1, "value": "giraffe"},
        {"count": 1, "value": "lion"},
        {"count": 4, "value": "zebra"},
    ]

    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=result,
        unexpected_index_column_names=unexpected_index_column_names,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["pk_1", "Count"]
    assert res.iloc[0, 0] == "3"
    assert res.iloc[1, 0] == "4"
    assert res.iloc[2, 0] == "5, 6, 7, 8"
    assert res.iloc[0, 1] == 1
    assert res.iloc[1, 1] == 1
    assert res.iloc[2, 1] == 4


def test_convert_unexpected_indices_to_df_multiple_with_truncation():
    result = [
        {"animals": "zebra", "pk_1": 0},
        {"animals": "zebra", "pk_1": 1},
        {"animals": "zebra", "pk_1": 2},
        {"animals": "zebra", "pk_1": 3},
        {"animals": "zebra", "pk_1": 4},
        {"animals": "zebra", "pk_1": 5},
        {"animals": "zebra", "pk_1": 6},
        {"animals": "zebra", "pk_1": 7},
        {"animals": "zebra", "pk_1": 8},
        {"animals": "zebra", "pk_1": 9},
        {"animals": "zebra", "pk_1": 10},
        {"animals": "zebra", "pk_1": 11},
        {"animals": "zebra", "pk_1": 12},
    ]
    unexpected_index_column_names = ["pk_1"]
    partial_unexpected_counts = [{"count": 13, "value": "zebra"}]
    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=result,
        unexpected_index_column_names=unexpected_index_column_names,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["pk_1", "Count"]
    assert res.iloc[0].tolist() == ["0, 1, 2, 3, 4, 5, 6, 7, 8, 9, ...", 13]


def test_convert_unexpected_indices_to_df_column_pair_expectation():
    partial_unexpected_counts = [{"value": ("eraser", "desk"), "count": 3}]
    unexpected_index_list = [
        {"ordered_item": "eraser", "received_item": "desk", "pk_2": "three"},
        {"ordered_item": "eraser", "received_item": "desk", "pk_2": "four"},
        {"ordered_item": "eraser", "received_item": "desk", "pk_2": "five"},
    ]
    unexpected_index_column_names = ["pk_2"]
    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=unexpected_index_list,
        unexpected_index_column_names=unexpected_index_column_names,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["pk_2", "Count"]
    assert res.index.to_list() == ["('eraser', 'desk')"] or res.index.to_list() == [
        "('desk', 'eraser')"
    ]
    assert res.iloc[0].tolist() == ["three, four, five", 3]


@pytest.mark.skipif(
    sys.version_info < (3, 9),
    reason="Failing on Python 3.7 and 3.8; blocking 0.15.48 release",
)
def test_convert_unexpected_indices_to_df_column_pair_expectation_no_id_pk():
    partial_unexpected_counts = [{"value": ("eraser", "desk"), "count": 3}]
    unexpected_index_list = [3, 4, 5]
    unexpected_list = [("eraser", "desk"), ("eraser", "desk"), ("eraser", "desk")]
    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=unexpected_index_list,
        unexpected_index_column_names=None,
        unexpected_list=unexpected_list,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["Index", "Count"]
    assert res.index.to_list() == ["('eraser', 'desk')"] or res.index.to_list() == [
        "('desk', 'eraser')"
    ]
    assert res.iloc[0].tolist() == ["3, 4, 5", 3]


def test_convert_unexpected_indices_to_df_multiple_column_expectation():
    partial_unexpected_counts = [
        {"value": (20, 20, 20), "count": 1},
        {"value": (30, 30, 30), "count": 1},
        {"value": (40, 40, 40), "count": 1},
        {"value": (50, 50, 50), "count": 1},
        {"value": (60, 60, 60), "count": 1},
    ]
    unexpected_index_list = [
        {"a": 20, "pk_1": 1, "b": 20, "c": 20},
        {"a": 30, "pk_1": 2, "b": 30, "c": 30},
        {"a": 40, "pk_1": 3, "b": 40, "c": 40},
        {"a": 50, "pk_1": 4, "b": 50, "c": 50},
        {"a": 60, "pk_1": 5, "b": 60, "c": 60},
    ]
    unexpected_index_column_names = ["pk_1"]
    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=unexpected_index_list,
        unexpected_index_column_names=unexpected_index_column_names,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["pk_1", "Count"]
    assert res.index.to_list() == [
        "('20', '20', '20')",
        "('30', '30', '30')",
        "('40', '40', '40')",
        "('50', '50', '50')",
        "('60', '60', '60')",
    ]
    assert res.iloc[0].tolist() == ["1", 1]
    assert res.iloc[1].tolist() == ["2", 1]
    assert res.iloc[2].tolist() == ["3", 1]
    assert res.iloc[3].tolist() == ["4", 1]
    assert res.iloc[4].tolist() == ["5", 1]


@pytest.mark.skipif(
    sys.version_info < (3, 9),
    reason="Failing on Python 3.7 and 3.8; blocking 0.15.48 release",
)
def test_convert_unexpected_indices_to_df_multiple_column_expectation_no_id_pk():
    partial_unexpected_counts = [
        {"value": (20, 20, 20), "count": 1},
        {"value": (30, 30, 30), "count": 1},
        {"value": (40, 40, 40), "count": 1},
        {"value": (50, 50, 50), "count": 1},
        {"value": (60, 60, 60), "count": 1},
    ]
    unexpected_list = [
        (20, 20, 20),
        (30, 30, 30),
        (40, 40, 40),
        (50, 50, 50),
        (60, 60, 60),
    ]
    unexpected_index_list = [1, 2, 3, 4, 5]
    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=unexpected_index_list,
        unexpected_index_column_names=None,
        unexpected_list=unexpected_list,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["Index", "Count"]
    assert res.index.to_list() == [
        "(20, 20, 20)",
        "(30, 30, 30)",
        "(40, 40, 40)",
        "(50, 50, 50)",
        "(60, 60, 60)",
    ]
    assert res.iloc[0].tolist() == ["1", 1]
    assert res.iloc[1].tolist() == ["2", 1]
    assert res.iloc[2].tolist() == ["3", 1]
    assert res.iloc[3].tolist() == ["4", 1]
    assert res.iloc[4].tolist() == ["5", 1]


def test_convert_unexpected_indices_to_df_actual_values():
    """
    Test if no index_column_names are specified, and we use Pandas' default unexpected indices
    """
    unexpected_index_list = [3, 4, 5]
    unexpected_list = ["giraffe", "lion", "zebra"]
    partial_unexpected_counts = [
        {"count": 1, "value": "giraffe"},
        {"count": 1, "value": "lion"},
        {"count": 4, "value": "zebra"},
    ]

    res = _convert_unexpected_indices_to_df(
        unexpected_index_list=unexpected_index_list,
        unexpected_list=unexpected_list,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    assert list(res) == ["Index", "Count"]
    assert res.iloc[0].tolist() == ["3", 1]
    assert res.iloc[1].tolist() == ["4", 1]
    assert res.iloc[2].tolist() == ["5", 1]


def test_truncate_list_of_indices():
    int_indices: List[Union[int, str]] = [4, 5, 6, 7]
    result: str = truncate_list_of_indices(indices=int_indices)
    assert result == "4, 5, 6, 7"

    result: str = truncate_list_of_indices(indices=int_indices, max_index=2)
    assert result == "4, 5, ..."

    str_indices: List[Union[int, str]] = ["four", "five", "six", "seven"]
    result: str = truncate_list_of_indices(indices=str_indices)
    assert result == "four, five, six, seven"

    result: str = truncate_list_of_indices(indices=str_indices, max_index=2)
    assert result == "four, five, ..."


def test_build_count_and_index_table():
    partial_unexpected_counts = [
        {"count": 1, "value": "giraffe"},
        {"count": 1, "value": "lion"},
        {"count": 1, "value": "zebra"},
    ]
    unexpected_index_list = [
        {"animals": "giraffe", "pk_1": 3, "pk_2": "three"},
        {"animals": "lion", "pk_1": 4, "pk_2": "four"},
        {"animals": "zebra", "pk_1": 5, "pk_2": "five"},
    ]
    unexpected_count = 3
    unexpected_index_column_names = ["pk_1", "pk_2"]

    header_row, table_rows = build_count_and_index_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_index_list=unexpected_index_list,
        unexpected_count=unexpected_count,
        unexpected_index_column_names=unexpected_index_column_names,
    )
    assert header_row == ["Unexpected Value", "Count", "pk_1", "pk_2"]
    assert table_rows == [
        ["giraffe", 1, "3", "three"],
        ["lion", 1, "4", "four"],
        ["zebra", 1, "5", "five"],
    ]


def test_build_count_and_index_table_sampled():
    partial_unexpected_counts = [
        {"count": 1, "value": "giraffe"},
        {"count": 1, "value": "lion"},
        {"count": 1, "value": "zebra"},
    ]
    unexpected_index_list = [
        {"animals": "giraffe", "pk_1": 3, "pk_2": "three"},
        {"animals": "lion", "pk_1": 4, "pk_2": "four"},
        {"animals": "zebra", "pk_1": 5, "pk_2": "five"},
    ]
    unexpected_count = 100
    unexpected_index_column_names = ["pk_1", "pk_2"]

    header_row, table_rows = build_count_and_index_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_index_list=unexpected_index_list,
        unexpected_count=unexpected_count,
        unexpected_index_column_names=unexpected_index_column_names,
    )
    assert header_row == ["Sampled Unexpected Values", "Count", "pk_1", "pk_2"]
    assert table_rows == [
        ["giraffe", 1, "3", "three"],
        ["lion", 1, "4", "four"],
        ["zebra", 1, "5", "five"],
    ]


def test_build_count_and_index_table_with_empty_string():
    partial_unexpected_counts = [
        {"count": 1, "value": ""},
    ]
    unexpected_index_list = [
        {"animals": "", "pk_1": 5, "pk_2": "five"},
    ]
    unexpected_count = 100
    unexpected_index_column_names = ["pk_1", "pk_2"]

    header_row, table_rows = build_count_and_index_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_index_list=unexpected_index_list,
        unexpected_count=unexpected_count,
        unexpected_index_column_names=unexpected_index_column_names,
    )
    assert header_row == ["Sampled Unexpected Values", "Count", "pk_1", "pk_2"]
    assert table_rows == [
        ["EMPTY", 1, "5", "five"],
    ]


def test_build_count_and_index_table_with_multi_column():
    partial_unexpected_counts = [
        {"value": (20, 20, 20), "count": 1},
        {"value": (30, 30, 30), "count": 1},
        {"value": (40, 40, 40), "count": 1},
        {"value": (50, 50, 50), "count": 1},
        {"value": (60, 60, 60), "count": 1},
    ]
    unexpected_index_list = [
        {"a": 20, "pk_1": 1, "b": 20, "c": 20},
        {"a": 30, "pk_1": 2, "b": 30, "c": 30},
        {"a": 40, "pk_1": 3, "b": 40, "c": 40},
        {"a": 50, "pk_1": 4, "b": 50, "c": 50},
        {"a": 60, "pk_1": 5, "b": 60, "c": 60},
    ]
    unexpected_count = 5
    unexpected_index_column_names = ["pk_1"]
    header_row, table_rows = build_count_and_index_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_index_list=unexpected_index_list,
        unexpected_count=unexpected_count,
        unexpected_index_column_names=unexpected_index_column_names,
    )
    assert header_row == ["Unexpected Value", "Count", "pk_1"]
    assert table_rows == [
        ["('20', '20', '20')", 1, "1"],
        ["('30', '30', '30')", 1, "2"],
        ["('40', '40', '40')", 1, "3"],
        ["('50', '50', '50')", 1, "4"],
        ["('60', '60', '60')", 1, "5"],
    ]


def test_build_count_and_index_table_with_column_pair():
    partial_unexpected_counts = [{"value": ("eraser", "desk"), "count": 3}]
    unexpected_index_list = [
        {"ordered_item": "eraser", "received_item": "desk", "pk_2": "three"},
        {"ordered_item": "eraser", "received_item": "desk", "pk_2": "four"},
        {"ordered_item": "eraser", "received_item": "desk", "pk_2": "five"},
    ]
    unexpected_count = 3
    unexpected_index_column_names = ["pk_2"]
    header_row, table_rows = build_count_and_index_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_index_list=unexpected_index_list,
        unexpected_count=unexpected_count,
        unexpected_index_column_names=unexpected_index_column_names,
    )
    assert header_row == ["Unexpected Value", "Count", "pk_2"]
    assert table_rows == [  # noqa: PLR1714 # FIXME CoP
        ["('desk', 'eraser')", 3, "three, four, five"]
    ] or table_rows == [["('eraser', 'desk')", 3, "three, four, five"]]


def test_build_count_and_index_table_with_null():
    partial_unexpected_counts = [
        {"count": 1, "value": None},
    ]
    unexpected_index_list = [
        {"animals": None, "pk_1": 5, "pk_2": "five"},
    ]
    unexpected_count = 100
    unexpected_index_column_names = ["pk_1", "pk_2"]

    header_row, table_rows = build_count_and_index_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_index_list=unexpected_index_list,
        unexpected_count=unexpected_count,
        unexpected_index_column_names=unexpected_index_column_names,
    )
    assert header_row == ["Sampled Unexpected Values", "Count", "pk_1", "pk_2"]
    assert table_rows == [
        ["null", 1, "5", "five"],
    ]


def test_build_count_table():
    partial_unexpected_counts = [
        {"count": 3, "value": "giraffe"},
        {"count": 2, "value": "lion"},
        {"count": 1, "value": "zebra"},
    ]
    unexpected_count = 6

    header_row, table_rows = build_count_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_count=unexpected_count,
    )
    assert header_row == ["Unexpected Value", "Count"]
    assert table_rows == [["giraffe", 3], ["lion", 2], ["zebra", 1]]


def test_build_count_table_sampled():
    partial_unexpected_counts = [
        {"count": 3, "value": "giraffe"},
        {"count": 2, "value": "lion"},
        {"count": 1, "value": "zebra"},
    ]
    unexpected_count = 100

    header_row, table_rows = build_count_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_count=unexpected_count,
    )
    assert header_row == ["Sampled Unexpected Values"]
    assert table_rows == [["giraffe"], ["lion"], ["zebra"]]


def test_build_count_table_with_empty_string():
    partial_unexpected_counts = [
        {"count": 1, "value": ""},
    ]
    unexpected_count = 100

    header_row, table_rows = build_count_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_count=unexpected_count,
    )
    assert header_row == ["Sampled Unexpected Values"]
    assert table_rows == [
        ["EMPTY"],
    ]


def test_build_count_table_with_null():
    partial_unexpected_counts = [
        {"count": 1, "value": None},
    ]
    unexpected_count = 100

    header_row, table_rows = build_count_table(
        partial_unexpected_counts=partial_unexpected_counts,
        unexpected_count=unexpected_count,
    )
    assert header_row == ["Sampled Unexpected Values"]
    assert table_rows == [["null"]]



import datetime as dt

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    SQL_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
)

PANDAS_DATA_SOURCES: list[DataSourceTestConfig] = [
    ds for ds in ALL_DATA_SOURCES if ds.label.startswith("pandas")
]
SPARK_DATA_SOURCES: list[DataSourceTestConfig] = [
    ds for ds in ALL_DATA_SOURCES if ds.label.startswith("spark")
]


@pytest.mark.unit
def test_parameterization():
    # Ensure that all data sources are covered
    assert len(PANDAS_DATA_SOURCES) + len(SPARK_DATA_SOURCES) + len(SQL_DATA_SOURCES) == len(
        ALL_DATA_SOURCES
    )


class TestNumericExpectationAgainstStrDataMisconfiguration:
    # Currently bugs with the following (not raising misconfiguration errors at all):
    #  - sqlite
    #  - databricks
    #  - mysql
    #  - spark-filesystem-csv

    _DATA = pd.DataFrame({"a": ["b", "c"]})
    _EXPECTATION = gxe.ExpectColumnStdevToBeBetween(
        column="a",
        min_value=0,
        max_value=1,
        strict_min=True,
        strict_max=True,
    )

    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES,
        data=_DATA,
    )
    def test_pandas(self, batch_for_datasource) -> None:
        self._assert_misconfiguration(
            batch_for_datasource=batch_for_datasource,
            exception_message="could not convert string to float",
        )

    @parameterize_batch_for_data_sources(
        data_source_configs=[BigQueryDatasourceTestConfig()],
        data=_DATA,
    )
    def test_bigquery(self, batch_for_datasource) -> None:
        self._assert_misconfiguration(
            batch_for_datasource=batch_for_datasource,
            exception_message="No matching signature for operator *",
        )

    @parameterize_batch_for_data_sources(
        data_source_configs=[MSSQLDatasourceTestConfig()],
        data=_DATA,
    )
    def test_mssql(self, batch_for_datasource) -> None:
        self._assert_misconfiguration(
            batch_for_datasource=batch_for_datasource,
            exception_message="Error converting data type varchar to float",
        )

    @parameterize_batch_for_data_sources(
        data_source_configs=[PostgreSQLDatasourceTestConfig()],
        data=_DATA,
    )
    def test_postgresql(self, batch_for_datasource) -> None:
        self._assert_misconfiguration(
            batch_for_datasource=batch_for_datasource,
            exception_message="operator does not exist: numeric * character varying",
        )

    @parameterize_batch_for_data_sources(
        data_source_configs=[SnowflakeDatasourceTestConfig()],
        data=_DATA,
    )
    def test_snowflake(self, batch_for_datasource) -> None:
        self._assert_misconfiguration(
            batch_for_datasource=batch_for_datasource,
            exception_message="Numeric value \\'b\\' is not recognized",
        )

    def _assert_misconfiguration(self, batch_for_datasource: Batch, exception_message: str) -> None:
        result = batch_for_datasource.validate(self._EXPECTATION)
        assert not result.success
        assert exception_message in str(result.exception_info)


class TestNonExistentColumnMisconfiguration:
    _DATA = pd.DataFrame({"a": [1, 2]})
    _EXPECTATION = gxe.ExpectColumnMedianToBeBetween(column="b", min_value=5, max_value=10)

    @parameterize_batch_for_data_sources(
        data_source_configs=[*PANDAS_DATA_SOURCES, *SQL_DATA_SOURCES],
        data=_DATA,
    )
    def test_pandas_and_sql(self, batch_for_datasource) -> None:
        self._assert_misconfiguration(
            batch_for_datasource=batch_for_datasource,
            exception_message='The column "b" in BatchData does not exist',
        )

    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=_DATA,
    )
    def test_spark(self, batch_for_datasource) -> None:
        self._assert_misconfiguration(
            batch_for_datasource=batch_for_datasource,
            exception_message="A column or function parameter with name `b` cannot be resolved",
        )

    def _assert_misconfiguration(self, batch_for_datasource: Batch, exception_message: str) -> None:
        result = batch_for_datasource.validate(self._EXPECTATION)
        assert not result.success
        assert exception_message in str(result.exception_info)


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2]}),
)
def test_datetime_expectation_against_numeric_data_misconfiguration(batch_for_datasource) -> None:
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column="a",
        min_value=dt.datetime(2024, 1, 1, tzinfo=dt.timezone.utc),
        max_value=dt.datetime(2024, 1, 2, tzinfo=dt.timezone.utc),
        strict_max=True,
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success
    assert "into datetime representation" in str(result.exception_info)


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2]}),
)
def test_column_min_max_mismatch_misconfiguration(batch_for_datasource) -> None:
    expectation = gxe.ExpectColumnValuesToBeBetween(column="a", min_value=2, max_value=1)
    result = batch_for_datasource.validate(expectation)
    assert not result.success
    assert "min_value cannot be greater than max_value" in str(result.exception_info)



from __future__ import annotations

from typing import Mapping

import pytest

from great_expectations.data_context.data_context.context_factory import set_context
from great_expectations.data_context.data_context.ephemeral_data_context import (
    EphemeralDataContext,
)
from great_expectations.data_context.store import ExpectationsStore
from great_expectations.data_context.store.checkpoint_store import CheckpointStore
from great_expectations.data_context.store.datasource_store import DatasourceStore
from great_expectations.data_context.types.base import (
    DataContextConfig,
    InMemoryStoreBackendDefaults,
    ProgressBarsConfig,
)
from great_expectations.datasource.fluent.sources import DataSourceManager
from great_expectations.exceptions.exceptions import StoreConfigurationError


class DatasourceStoreSpy(DatasourceStore):
    def __init__(self) -> None:
        self.save_count = 0
        super().__init__()

    def set(self, key, value, **kwargs):
        ret = super().set(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret


class ExpectationsStoreSpy(ExpectationsStore):
    def __init__(self) -> None:
        self.save_count = 0
        super().__init__()

    def add(self, key, value, **kwargs):
        ret = super().add(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def update(self, key, value, **kwargs):
        ret = super().update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def add_or_update(self, key, value, **kwargs):
        ret = super().add_or_update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret


class CheckpointStoreSpy(CheckpointStore):
    STORE_NAME = "checkpoint_store"

    def __init__(self) -> None:
        self.save_count = 0
        super().__init__(store_name=CheckpointStoreSpy.STORE_NAME)

    def add(self, key, value, **kwargs):
        ret = super().add(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def update(self, key, value, **kwargs):
        ret = super().update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def add_or_update(self, key, value, **kwargs):
        ret = super().add_or_update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret


class EphemeralDataContextSpy(EphemeralDataContext):
    """
    Simply wraps around EphemeralDataContext but keeps tabs on specific method calls around state management.
    """  # noqa: E501 # FIXME CoP

    def __init__(
        self,
        project_config: DataContextConfig,
    ) -> None:
        # expectation store is required for initializing the base DataContext
        self._expectations_store = ExpectationsStoreSpy()
        self._checkpoint_store = CheckpointStoreSpy()
        super().__init__(project_config)
        self.save_count = 0
        self._datasource_store = DatasourceStoreSpy()

    @property
    def datasource_store(self):
        return self._datasource_store

    @property
    def expectations_store(self):
        return self._expectations_store

    @property
    def checkpoint_store(self):
        return self._checkpoint_store

    def _save_project_config(self):
        """
        No-op our persistence mechanism but increment an internal counter to ensure it was used.
        """
        self.save_count += 1


BLOCK_CONFIG_DATASOURCE_NAME = "my_pandas_datasource"


@pytest.fixture
def in_memory_data_context(
    fluent_datasource_config: dict,
) -> EphemeralDataContextSpy:
    config = DataContextConfig(
        store_backend_defaults=InMemoryStoreBackendDefaults(),
    )
    context = EphemeralDataContextSpy(project_config=config)
    ds_type = DataSourceManager.type_lookup[fluent_datasource_config["type"]]
    fluent_datasources = {
        fluent_datasource_config["name"]: ds_type(**fluent_datasource_config),
    }
    context.data_sources.all().update(fluent_datasources)
    set_context(context)
    return context


@pytest.mark.unit
def test_add_store(in_memory_data_context: EphemeralDataContextSpy):
    context = in_memory_data_context

    num_stores_before = len(context.stores)
    num_store_configs_before = len(context.config.stores)

    context.add_store(
        name="my_new_store",
        config={
            "module_name": "great_expectations.data_context.store",
            "class_name": "ExpectationsStore",
        },
    )

    num_stores_after = len(context.stores)
    num_store_configs_after = len(context.config.stores)

    assert num_stores_after == num_stores_before + 1
    assert num_store_configs_after == num_store_configs_before + 1
    assert context.save_count == 1


@pytest.mark.unit
def test_delete_store_success(in_memory_data_context: EphemeralDataContextSpy):
    context = in_memory_data_context

    num_stores_before = len(context.stores)
    num_store_configs_before = len(context.config.stores)

    context.delete_store("checkpoint_store")  # We know this to be a default name

    num_stores_after = len(context.stores)
    num_store_configs_after = len(context.config.stores)

    assert num_stores_after == num_stores_before - 1
    assert num_store_configs_after == num_store_configs_before - 1
    assert context.save_count == 1


@pytest.mark.unit
def test_delete_store_failure(in_memory_data_context: EphemeralDataContextSpy):
    context = in_memory_data_context

    num_stores_before = len(context.stores)
    num_store_configs_before = len(context.config.stores)

    with pytest.raises(StoreConfigurationError):
        context.delete_store("my_fake_store_name")

    num_stores_after = len(context.stores)
    num_store_configs_after = len(context.config.stores)

    assert num_stores_after == num_stores_before
    assert num_store_configs_after == num_store_configs_before
    assert context.save_count == 0


@pytest.mark.unit
@pytest.mark.parametrize(
    "config",
    [
        pytest.param(
            DataContextConfig(progress_bars=ProgressBarsConfig(globally=True)),
            id="DataContextConfig",
        ),
        pytest.param({"progress_bars": ProgressBarsConfig(globally=True)}, id="Mapping"),
    ],
)
def test_update_project_config(
    in_memory_data_context: EphemeralDataContextSpy, config: DataContextConfig | Mapping
):
    context = in_memory_data_context

    assert context.progress_bars is None

    context.update_project_config(config)

    assert context.progress_bars["globally"] is True



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

# TODO: Full coverage to replace test_expect_column_kl_divergence_to_be_less_than.py

COL_NAME = "my_col"

DATA = pd.DataFrame(
    {
        COL_NAME: ["A"] * 5 + ["B"] * 3 + ["C"] * 2,
    }
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnKLDivergenceToBeLessThan(
        column=COL_NAME,
        partition_object={"weights": [0.5, 0.3, 0.2], "values": ["A", "B", "C"]},
        threshold=0.01,
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "observed_value": 0.0,
        "details": {
            "observed_partition": {"values": ["A", "B", "C"], "weights": [0.5, 0.3, 0.2]},
            "expected_partition": {"values": ["A", "B", "C"], "weights": [0.5, 0.3, 0.2]},
        },
    }


@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnKLDivergenceToBeLessThan(
        column=COL_NAME,
        partition_object={
            "weights": [0.3333333333333333, 0.3333333333333333, 0.3333333333333333],
            "values": ["A", "B", "C"],
        },
        threshold=0.01,
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param({"weights": [0.5, 0.3, 0.2], "values": ["A", "B", "C"]}, True, id="success"),
        pytest.param(
            {
                "weights": [0.3333333333333333, 0.3333333333333333, 0.3333333333333333],
                "values": ["A", "B", "C"],
            },
            False,
            id="failure",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_partition_object_(
    batch_for_datasource: Batch, suite_param_value: dict, expected_result: bool
) -> None:
    suite_param_key = "expect_column_kl_divergence_to_be_less_than"
    expectation = gxe.ExpectColumnKLDivergenceToBeLessThan(
        column=COL_NAME,
        partition_object={"$PARAMETER": suite_param_key},
        threshold=0.01,
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(0.01, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_threshold_(
    batch_for_datasource: Batch, suite_param_value: float, expected_result: bool
) -> None:
    suite_param_key = "expect_column_kl_divergence_to_be_less_than"
    expectation = gxe.ExpectColumnKLDivergenceToBeLessThan(
        column=COL_NAME,
        partition_object={"weights": [0.5, 0.3, 0.2], "values": ["A", "B", "C"]},
        threshold={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(0, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_internal_weight_holdout_(
    batch_for_datasource: Batch, suite_param_value: float, expected_result: bool
) -> None:
    suite_param_key = "expect_column_kl_divergence_to_be_less_than"
    expectation = gxe.ExpectColumnKLDivergenceToBeLessThan(
        column=COL_NAME,
        partition_object={"weights": [0.5, 0.3, 0.2], "values": ["A", "B", "C"]},
        threshold=0.01,
        internal_weight_holdout={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(0.1, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_tail_weight_holdout_(
    batch_for_datasource: Batch, suite_param_value: float, expected_result: bool
) -> None:
    suite_param_key = "expect_column_kl_divergence_to_be_less_than"
    expectation = gxe.ExpectColumnKLDivergenceToBeLessThan(
        column=COL_NAME,
        partition_object={"weights": [0.5, 0.3, 0.2], "values": ["A", "B", "C"]},
        threshold=0.01,
        tail_weight_holdout={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_bucketize_data_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "expect_column_kl_divergence_to_be_less_than"
    expectation = gxe.ExpectColumnKLDivergenceToBeLessThan(
        column=COL_NAME,
        partition_object={"weights": [0.5, 0.3, 0.2], "values": ["A", "B", "C"]},
        threshold=0.01,
        bucketize_data={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



import os
import shutil
from typing import Dict

import pytest

from great_expectations.data_context import get_context
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.store import ExpectationsStore, ValidationResultsStore
from great_expectations.data_context.util import (
    file_relative_path,
    instantiate_class_from_config,
)

# module level markers
pytestmark = pytest.mark.filesystem


def assert_how_to_buttons(
    context,
    index_page_locator_info: str,
    index_links_dict: Dict,
    show_how_to_buttons=True,
):
    """Helper function to assert presence or non-presence of how-to buttons and related content in various
    Data Docs pages.
    """  # noqa: E501 # FIXME CoP

    # these are simple checks for presence of certain page elements
    show_walkthrough_button = "Show Walkthrough"
    walkthrough_modal = "Great Expectations Walkthrough"
    cta_footer = "To continue exploring Great Expectations check out one of these tutorials..."
    how_to_edit_suite_button = "How to Edit This Suite"
    how_to_edit_suite_modal = "How to Edit This Expectation Suite"
    action_card = "Actions"

    how_to_page_elements_dict = {
        "index_pages": [show_walkthrough_button, walkthrough_modal, cta_footer],
        "expectation_suites": [
            how_to_edit_suite_button,
            how_to_edit_suite_modal,
            show_walkthrough_button,
            walkthrough_modal,
        ],
        "validation_results": [
            how_to_edit_suite_button,
            how_to_edit_suite_modal,
            show_walkthrough_button,
            walkthrough_modal,
        ],
        "profiling_results": [action_card, show_walkthrough_button, walkthrough_modal],
    }

    data_docs_site_dir = os.path.join(  # noqa: PTH118 # FIXME CoP
        context._context_root_directory,
        context._project_config.data_docs_sites["local_site"]["store_backend"]["base_directory"],
    )

    page_paths_dict = {
        "index_pages": [index_page_locator_info[7:]],
        "expectation_suites": [
            os.path.join(data_docs_site_dir, link_dict["filepath"])  # noqa: PTH118 # FIXME CoP
            for link_dict in index_links_dict.get("expectations_links", [])
        ],
        "validation_results": [
            os.path.join(data_docs_site_dir, link_dict["filepath"])  # noqa: PTH118 # FIXME CoP
            for link_dict in index_links_dict.get("validations_links", [])
        ],
        "profiling_results": [
            os.path.join(data_docs_site_dir, link_dict["filepath"])  # noqa: PTH118 # FIXME CoP
            for link_dict in index_links_dict.get("profiling_links", [])
        ],
    }

    for page_type, page_paths in page_paths_dict.items():
        for page_path in page_paths:
            with open(page_path) as f:
                page = f.read()
                for how_to_element in how_to_page_elements_dict[page_type]:
                    if show_how_to_buttons:
                        assert how_to_element in page
                    else:
                        assert how_to_element not in page


def test_site_builder_with_custom_site_section_builders_config(tmp_path_factory):
    """Test that site builder can handle partially specified custom site_section_builders config"""
    base_dir = str(tmp_path_factory.mktemp("project_dir"))
    project_dir = os.path.join(base_dir, "project_path")  # noqa: PTH118 # FIXME CoP
    os.mkdir(project_dir)  # noqa: PTH102 # FIXME CoP

    # fixture config swaps site section builder source stores and specifies custom run_name_filters
    shutil.copy(
        file_relative_path(
            __file__, "../test_fixtures/great_expectations_custom_local_site_config.yml"
        ),
        str(os.path.join(project_dir, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
    )
    context = get_context(context_root_dir=project_dir)
    local_site_config = context._project_config.data_docs_sites.get("local_site")

    module_name = "great_expectations.render.renderer.site_builder"
    site_builder = instantiate_class_from_config(
        config=local_site_config,
        runtime_environment={
            "data_context": context,
            "root_directory": context.root_directory,
            "site_name": "local_site",
        },
        config_defaults={"module_name": module_name},
    )
    site_section_builders = site_builder.site_section_builders

    expectations_site_section_builder = site_section_builders["expectations"]
    assert isinstance(expectations_site_section_builder.source_store, ValidationResultsStore)

    validations_site_section_builder = site_section_builders["validations"]
    assert isinstance(validations_site_section_builder.source_store, ExpectationsStore)
    assert validations_site_section_builder.run_name_filter == {
        "not_equals": "custom_validations_filter"
    }

    profiling_site_section_builder = site_section_builders["profiling"]
    assert isinstance(validations_site_section_builder.source_store, ExpectationsStore)
    assert profiling_site_section_builder.run_name_filter == {"equals": "custom_profiling_filter"}



from typing import List

import pytest

# module level markers
pytestmark = pytest.mark.filesystem


def test_list_datasources_base_data_context_no_datasources(empty_data_context) -> None:
    """What does this test and why?

    When listing datasources, we want to omit the name and id fields. This test uses DataContext.
    """

    context = empty_data_context

    # no datasources

    observed: List[dict] = context.list_datasources()

    expected: List[dict] = []

    assert observed == expected


def test_list_datasources_base_data_context_one_datasource(empty_data_context) -> None:
    """What does this test and why?

    When listing datasources, we want to omit the name and id fields. This test uses DataContext.
    """

    context = empty_data_context

    # one datasource

    datasource = context.data_sources.add_pandas("my_data_source")

    observed = context.list_datasources()

    assert observed == [datasource]


def test_list_datasources_base_data_context_two_datasources(empty_data_context) -> None:
    """What does this test and why?

    When listing datasources, we want to omit the name and id fields. This test uses DataContext.
    """

    context = empty_data_context

    data_source_a = context.data_sources.add_pandas("my_data_source")
    data_source_b = context.data_sources.add_pandas("your_data_source")

    observed = context.list_datasources()

    assert observed == [data_source_a, data_source_b]



from __future__ import annotations

import copy
import inspect
import logging
import pathlib
import re
import warnings
from typing import Any, Optional
from urllib.parse import urlparse

import pyparsing as pp

from great_expectations.alias_types import PathStr  # noqa: TC001 # FIXME CoP
from great_expectations.exceptions import StoreConfigurationError
from great_expectations.types import safe_deep_copy
from great_expectations.util import load_class, verify_dynamic_loading_support

try:
    import sqlalchemy as sa  # noqa: TID251 # FIXME CoP
except ImportError:
    sa = None  # type: ignore[assignment] # FIXME CoP

logger = logging.getLogger(__name__)


# TODO: Rename config to constructor_kwargs and config_defaults -> constructor_kwarg_default
# TODO: Improve error messages in this method. Since so much of our workflow is config-driven, this will be a *super* important part of DX.  # noqa: E501 # FIXME CoP
def instantiate_class_from_config(  # noqa: C901 # FIXME CoP
    config, runtime_environment, config_defaults=None
):
    """Build a GX class from configuration dictionaries."""

    if config_defaults is None:
        config_defaults = {}

    config = copy.deepcopy(config)

    module_name = config.pop("module_name", None)
    if module_name is None:
        try:
            module_name = config_defaults.pop("module_name")
        except KeyError:
            raise KeyError(  # noqa: TRY003 # FIXME CoP
                f"Neither config : {config} nor config_defaults : {config_defaults} contains a module_name key."  # noqa: E501 # FIXME CoP
            )
    else:
        # Pop the value without using it, to avoid sending an unwanted value to the config_class
        config_defaults.pop("module_name", None)

    logger.debug(f"(instantiate_class_from_config) module_name -> {module_name}")
    verify_dynamic_loading_support(module_name=module_name)

    class_name = config.pop("class_name", None)
    if class_name is None:
        logger.warning(
            "Instantiating class from config without an explicit class_name is dangerous. Consider adding "  # noqa: E501 # FIXME CoP
            f"an explicit class_name for {config.get('name')}"
        )
        try:
            class_name = config_defaults.pop("class_name")
        except KeyError:
            raise KeyError(  # noqa: TRY003 # FIXME CoP
                f"Neither config : {config} nor config_defaults : {config_defaults} contains a class_name key."  # noqa: E501 # FIXME CoP
            )
    else:
        # Pop the value without using it, to avoid sending an unwanted value to the config_class
        config_defaults.pop("class_name", None)

    class_ = load_class(class_name=class_name, module_name=module_name)

    config_with_defaults = copy.deepcopy(config_defaults)
    config_with_defaults.update(config)
    if runtime_environment is not None:
        # If there are additional kwargs available in the runtime_environment requested by a
        # class to be instantiated, provide them
        argspec = inspect.getfullargspec(class_.__init__)[0][1:]

        missing_args = set(argspec) - set(config_with_defaults.keys())
        config_with_defaults.update(
            {
                missing_arg: runtime_environment[missing_arg]
                for missing_arg in missing_args
                if missing_arg in runtime_environment
            }
        )
        # Add the entire runtime_environment as well if it's requested
        if "runtime_environment" in missing_args:
            config_with_defaults.update({"runtime_environment": runtime_environment})

    try:
        class_instance = class_(**config_with_defaults)
    except TypeError as e:
        raise TypeError(
            f"Couldn't instantiate class: {class_name} with config: \n\t{format_dict_for_error_message(config_with_defaults)}\n \n"  # noqa: E501 # FIXME CoP
            + str(e)
        )

    return class_instance


def format_dict_for_error_message(dict_):
    # TODO : Tidy this up a bit. Indentation isn't fully consistent.

    return "\n\t".join("\t\t".join((str(key), str(dict_[key]))) for key in dict_)


def file_relative_path(
    source_path: PathStr,
    relative_path: PathStr,
    strict: bool = True,
) -> str:
    """
    This function is useful when one needs to load a file that is
    relative to the position of the current file. (Such as when
    you encode a configuration file path in source file and want
    in runnable in any current working directory)

    It is meant to be used like the following:
    file_relative_path(__file__, 'path/relative/to/file')

    This has been modified from Dagster's utils:
    H/T https://github.com/dagster-io/dagster/blob/8a250e9619a49e8bff8e9aa7435df89c2d2ea039/python_modules/dagster/dagster/utils/__init__.py#L34
    """
    dir_path = pathlib.Path(source_path).parent
    abs_path = dir_path.joinpath(relative_path).resolve(strict=strict)
    return str(abs_path)


def parse_substitution_variable(substitution_variable: str) -> Optional[str]:
    """
    Parse and check whether the string contains a substitution variable of the case insensitive form ${SOME_VAR} or $SOME_VAR
    Args:
        substitution_variable: string to be parsed

    Returns:
        string of variable name e.g. SOME_VAR or None if not parsable. If there are multiple substitution variables this currently returns the first e.g. $SOME_$TRING -> $SOME_
    """  # noqa: E501 # FIXME CoP
    substitution_variable_name = pp.Word(pp.alphanums + "_").setResultsName(
        "substitution_variable_name"
    )
    curly_brace_parser = "${" + substitution_variable_name + "}"
    non_curly_brace_parser = "$" + substitution_variable_name
    both_parser = curly_brace_parser | non_curly_brace_parser
    try:
        parsed_substitution_variable = both_parser.parseString(substitution_variable)
        return parsed_substitution_variable.substitution_variable_name
    except pp.ParseException:
        return None


class PasswordMasker:
    """
    Used to mask passwords in Datasources. Does not mask sqlite urls.

    Example usage
    masked_db_url = PasswordMasker.mask_db_url(url)
    where url = "postgresql+psycopg2://username:password@host:65432/database"
    and masked_url = "postgresql+psycopg2://username:***@host:65432/database"

    """

    MASKED_PASSWORD_STRING = "***"

    # values with the following keys will be processed with cls.mask_db_url:
    URL_KEYS = {"conn_str", "connection_string", "url"}

    # values with these keys will be directly replaced with cls.MASKED_PASSWORD_STRING:
    PASSWORD_KEYS = {"access_token", "password"}

    @classmethod
    def mask_db_url(cls, url: str, use_urlparse: bool = False, **kwargs) -> str:
        """
        Mask password in database url unless it is a substitution string, e.g. ConfigStr.
        Uses sqlalchemy engine parsing if sqlalchemy is installed, otherwise defaults to using urlparse from the stdlib which does not handle kwargs.
        Args:
            url: Database url e.g. "postgresql+psycopg2://username:password@host:65432/database"
            use_urlparse: Skip trying to parse url with sqlalchemy and use urlparse
            **kwargs: passed to create_engine()

        Returns:
            url with password masked e.g. "postgresql+psycopg2://username:***@host:65432/database"
        """  # noqa: E501 # FIXME CoP

        from great_expectations.datasource.fluent.config_str import ConfigStr

        is_config_str = ConfigStr.str_contains_config_template(url)

        if url.startswith("DefaultEndpointsProtocol"):
            return cls._obfuscate_azure_blobstore_connection_string(url)
        elif is_config_str:
            return url
        elif sa is not None and use_urlparse is False:
            try:
                engine = sa.create_engine(url, **kwargs)
                return engine.url.__repr__()
            # Account for the edge case where we have SQLAlchemy in our env but haven't installed the appropriate dialect to match the input URL  # noqa: E501 # FIXME CoP
            except Exception as e:
                logger.warning(
                    f"Something went wrong when trying to use SQLAlchemy to obfuscate URL: {e}"
                )
        else:
            warnings.warn(
                "SQLAlchemy is not installed, using urlparse to mask database url password which ignores **kwargs."  # noqa: E501 # FIXME CoP
            )
        return cls._mask_db_url_no_sa(url=url)

    @classmethod
    def _obfuscate_azure_blobstore_connection_string(cls, url: str) -> str:
        # Parse Azure Connection Strings
        azure_conn_str_re = re.compile(
            "(DefaultEndpointsProtocol=(http|https));(AccountName=([a-zA-Z0-9]+));(AccountKey=)(.+);(EndpointSuffix=([a-zA-Z\\.]+))"
        )
        try:
            matched: re.Match[str] | None = azure_conn_str_re.match(url)
            if not matched:
                raise StoreConfigurationError(  # noqa: TRY003, TRY301 # FIXME CoP
                    f"The URL for the Azure connection-string, was not configured properly. Please check and try again: {url} "  # noqa: E501 # FIXME CoP
                )
            res = f"DefaultEndpointsProtocol={matched.group(2)};AccountName={matched.group(4)};AccountKey=***;EndpointSuffix={matched.group(8)}"  # noqa: E501 # FIXME CoP
            return res
        except Exception as e:
            raise StoreConfigurationError(  # noqa: TRY003 # FIXME CoP
                f"Something went wrong when trying to obfuscate URL for Azure connection-string. Please check your configuration: {e}"  # noqa: E501 # FIXME CoP
            )

    @classmethod
    def _mask_db_url_no_sa(cls, url: str) -> str:
        # oracle+cx_oracle does not parse well using urlparse, parse as oracle then swap back
        replace_prefix = None
        if url.startswith("oracle+cx_oracle"):
            replace_prefix = {"original": "oracle+cx_oracle", "temporary": "oracle"}
            url = url.replace(replace_prefix["original"], replace_prefix["temporary"])

        parsed_url = urlparse(url)

        # Do not parse sqlite
        if parsed_url.scheme == "sqlite":
            return url

        colon = ":" if parsed_url.port is not None else ""
        masked_url = (
            f"{parsed_url.scheme}://{parsed_url.username}:{cls.MASKED_PASSWORD_STRING}"
            f"@{parsed_url.hostname}{colon}{parsed_url.port or ''}{parsed_url.path or ''}"
        )

        if replace_prefix is not None:
            masked_url = masked_url.replace(replace_prefix["temporary"], replace_prefix["original"])

        return masked_url

    @classmethod
    def sanitize_config(cls, config: dict) -> dict:  # noqa: C901 #  too complex
        """
        Mask sensitive fields in a Dict.
        """

        # be defensive, since it would be logical to expect this method works with DataContextConfig
        if not isinstance(config, dict):
            raise TypeError(
                "PasswordMasker.sanitize_config expects param `config` "
                + f"to be of type Dict, not of type {type(config)}"
            )

        config_copy = safe_deep_copy(config)  # be immutable

        def recursive_cleaner_method(config: Any) -> None:
            if isinstance(config, dict):
                for key, val in config.items():
                    if not isinstance(val, str):
                        recursive_cleaner_method(val)
                    elif key in cls.URL_KEYS:
                        config[key] = cls.mask_db_url(val)
                    elif key in cls.PASSWORD_KEYS:
                        config[key] = cls.MASKED_PASSWORD_STRING
                    else:
                        pass  # this string is not sensitive
            elif isinstance(config, list):
                for val in config:
                    recursive_cleaner_method(val)

        recursive_cleaner_method(config_copy)  # Perform anonymization in place

        return config_copy



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

from great_expectations.exceptions import ExecutionEngineError
from great_expectations.self_check.util import (
    _check_if_valid_dataset_name,
    generate_dataset_name_from_expectation_name,
)

# module level markers
pytestmark = pytest.mark.unit


@pytest.mark.parametrize(
    "dataset,expectation_name,index,sub_index,expected_output,expectation",
    [
        pytest.param(
            {"dataset_name": "i_am_a_dataset"},
            "expect_things",
            1,
            None,
            "i_am_a_dataset",
            does_not_raise(),
            id="defined_in_dataset_dict",
        ),
        pytest.param(
            {},
            "expect_things",
            1,
            None,
            "expect_things_dataset_1",
            does_not_raise(),
            id="expectation_name_and_index",
        ),
        pytest.param(
            {},
            "expect_things",
            1,
            2,
            "expect_things_dataset_1_2",
            does_not_raise(),
            id="expectation_name_and_sub_index",
        ),
        pytest.param(
            {},
            "expect_many_many_many_many_many_many_many_many_many_many_things",
            1,
            None,
            "y_many_many_many_many_many_many_many_many_many_things_dataset_1",
            does_not_raise(),
            id="expection_name_truncated_and_underscore_removed",
        ),
        pytest.param(
            {},
            "i*am*not*valid",
            1,
            None,
            "",
            pytest.raises(ExecutionEngineError),
            id="expection_name_truncated_and_underscore_removed",
        ),
    ],
)
def test_generate_table_name_with_expectation(
    dataset: dict,
    expectation_name: str,
    expected_output: str,
    index: int,
    sub_index: int | None,
    expectation,
):
    """Test for helper method that automatically generates table name for tests
    Args:
        dataset (dict): dictionary of test configuration.
        expectation_name (str): name of Expectation under test.
        expected_output (str): what should the generated table name be?
        index (int): index of current dataset
        sub_index (int): optional sub_index if there dataset is part of a list.
    """
    with expectation:
        assert (
            generate_dataset_name_from_expectation_name(
                dataset=dataset,
                expectation_type=expectation_name,
                index=index,
                sub_index=sub_index,
            )
            == expected_output
        )


@pytest.mark.parametrize(
    "dataset_name,expected_output,expectation",
    [
        pytest.param(
            "i_am_a_dataset",
            "i_am_a_dataset",
            does_not_raise(),
            id="defined_in_dataset_dict",
        ),
        pytest.param(
            "expect_many_many_many_many_many_many_many_many_many_many_many_many_many_many_things",
            "y_many_many_many_many_many_many_many_many_many_many_many_things",
            does_not_raise(),
            id="expection_name_truncated_and_underscore_removed",
        ),
        pytest.param(
            "i*am*not*valid",
            "",
            pytest.raises(ExecutionEngineError),
            id="expection_name_truncated_and_underscore_removed",
        ),
        pytest.param(
            "_________i_have_too_many_underscores",
            "i_have_too_many_underscores",
            does_not_raise(),
            id="beginning_underscores_removed",
        ),
    ],
)
def test_check_if_valid_dataset_name(dataset_name: str, expected_output: str, expectation):
    """Test for helper method that ensures table names are valid for tests
    Args:
        dataset_name(str): candidate dataset_name.
        expected_output (str): what should the final table name be?
    """
    with expectation:
        assert (
            _check_if_valid_dataset_name(
                dataset_name=dataset_name,
            )
            == expected_output
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # pydantic isinstance
)
from great_expectations.expectations.expectation import (
    COLUMN_DESCRIPTION,
    ColumnAggregateExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.render import (
    LegacyDescriptiveRendererType,
    LegacyRendererType,
    RenderedStringTemplateContent,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    handle_strict_min_max,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the proportion of non-null values to be between a minimum value and a maximum value."
)
MIN_VALUE_DESCRIPTION = (
    "The minimum proportion of non-null values (proportions are on the range 0 to 1)."
)
MAX_VALUE_DESCRIPTION = (
    "The maximum proportion of non-null values (proportions are on the range 0 to 1)."
)
STRICT_MIN_DESCRIPTION = (
    "If True, the minimum proportion of non-null values must be strictly larger than min_value."
)
STRICT_MAX_DESCRIPTION = (
    "If True, the maximum proportion of non-null values must be strictly smaller than max_value."
)
DATA_QUALITY_ISSUES = [DataQualityIssues.COMPLETENESS.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectColumnProportionOfNonNullValuesToBeBetween(ColumnAggregateExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    For example, in a column containing [1, 2, None, 3, None, None, 4, 4, 4, 4], there are \
    7 non-null values and 10 total values for a proportion of 0.7.

    ExpectColumnProportionOfNonNullValuesToBeBetween is a \
    Column Aggregate Expectation.

    Column Aggregate Expectations are one of the most common types of Expectation.
    They are evaluated for a single column, and produce an aggregate Metric, such as a \
    mean, standard deviation, number of unique values, column type, etc.
    If that Metric meets the conditions you set, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        min_value (float or None): \
           {MIN_VALUE_DESCRIPTION}
        max_value (float or None): \
            {MAX_VALUE_DESCRIPTION}
        strict_min (boolean): \
            {STRICT_MIN_DESCRIPTION} default=False
        strict_max (boolean): \
            {STRICT_MAX_DESCRIPTION} default=False

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the \
            output without modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to \
        result_format, catch_exceptions, and meta.

    Notes:
        * min_value and max_value are both inclusive unless \
          strict_min or strict_max are set to True.
        * If min_value is None, then max_value is treated as an upper bound
        * If max_value is None, then min_value is treated as a lower bound
        * observed_value field in the result object is customized for this expectation to be \
          a float representing the proportion of non-null values in the column

    See Also:
        [ExpectColumnProportionOfUniqueValuesToBeBetween](https://greatexpectations.io/expectations/expect_column_proportion_of_unique_values_to_be_between)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	"aaa"   1
            1 	"abb"   None
            2 	"acc"   1
            3   None    3

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnProportionOfNonNullValuesToBeBetween(
                    column="test",
                    min_value=0,
                    max_value=0.8
                )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 0.75
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnProportionOfNonNullValuesToBeBetween(
                    column="test2",
                    min_value=0.3,
                    max_value=0.5,
                    strict_min=False,
                    strict_max=True
                )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 0.75
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """

    min_value: Optional[Union[float, SuiteParameterDict]] = pydantic.Field(
        default=None, description=MIN_VALUE_DESCRIPTION
    )
    max_value: Optional[Union[float, SuiteParameterDict]] = pydantic.Field(
        default=None, description=MAX_VALUE_DESCRIPTION
    )
    strict_min: bool = pydantic.Field(default=False, description=STRICT_MIN_DESCRIPTION)
    strict_max: bool = pydantic.Field(default=False, description=STRICT_MAX_DESCRIPTION)

    library_metadata = {
        "maturity": "production",
        "tags": ["core expectation", "column aggregate expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    _library_metadata = library_metadata

    metric_dependencies = ("column.non_null_proportion",)
    success_keys = (
        "min_value",
        "max_value",
        "strict_min",
        "strict_max",
    )

    args_keys = (
        "column",
        "min_value",
        "max_value",
        "strict_min",
        "strict_max",
    )

    class Config:
        title = "Expect column proportion of non-null values to be between"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnProportionOfNonNullValuesToBeBetween]
        ) -> None:
            ColumnAggregateExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    def _get_min_max_string(cls, renderer_configuration: RendererConfiguration) -> str:
        params = renderer_configuration.params

        if not params.min_value and not params.max_value:
            return "may have any proportion of non-null values."
        else:
            at_least_str = "greater than or equal to"
            if params.strict_min:
                at_least_str = cls._get_strict_min_string(
                    renderer_configuration=renderer_configuration
                )
            at_most_str = "less than or equal to"
            if params.strict_max:
                at_most_str = cls._get_strict_max_string(
                    renderer_configuration=renderer_configuration
                )
            if not params.min_value:
                return f"proportion of non-null values must be {at_most_str} $max_value."
            elif not params.max_value:
                return f"proportion of non-null values must be {at_least_str} $min_value."
            elif params.min_value.value != params.max_value.value:
                return (
                    f"proportion of non-null values must be {at_least_str} $min_value "
                    f"and {at_most_str} $max_value."
                )
            else:
                return "proportion of non-null values must be exactly $min_value."

    @classmethod
    @override
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("min_value", [RendererValueType.NUMBER, RendererValueType.DATETIME]),
            ("max_value", [RendererValueType.NUMBER, RendererValueType.DATETIME]),
            ("strict_min", RendererValueType.BOOLEAN),
            ("strict_max", RendererValueType.BOOLEAN),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        template_str = cls._get_min_max_string(renderer_configuration)

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @override
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> list[RenderedStringTemplateContent]:
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = (
            substitute_none_for_missing(
                configuration.kwargs,
                [
                    "column",
                    "min_value",
                    "max_value",
                    "row_condition",
                    "condition_parser",
                    "strict_min",
                    "strict_max",
                ],
            )
            if configuration
            else {}
        )

        if params["min_value"] is None and params["max_value"] is None:
            template_str = "may have any proportion of non-null values."
        else:
            at_least_str, at_most_str = handle_strict_min_max(params)
            if params["min_value"] is None:
                template_str = f"proportion of non-null values must be {at_most_str} $max_value."
            elif params["max_value"] is None:
                template_str = f"proportion of non-null values must be {at_least_str} $min_value."
            elif params["min_value"] != params["max_value"]:
                template_str = (
                    f"proportion of non-null values must be {at_least_str} $min_value "
                    f"and {at_most_str} $max_value."
                )
            else:
                template_str = "proportion of non-null values must be exactly $min_value."

        if include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            ),
        ]

    @classmethod
    @renderer(
        renderer_type=LegacyDescriptiveRendererType.COLUMN_PROPERTIES_TABLE_DISTINCT_PERCENT_ROW
    )
    def _descriptive_column_properties_table_distinct_percent_row_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> list[Union[RenderedStringTemplateContent, str]]:
        assert result, "Must pass in result."
        observed_value = result.result["observed_value"]
        template_string_object = RenderedStringTemplateContent(
            content_block_type="string_template",
            string_template={
                "template": "Non-null (%)",
                "tooltip": {"content": "expect_column_proportion_of_non_null_values_to_be_between"},
            },
        )
        if not observed_value:
            return [template_string_object, "--"]
        else:
            return [template_string_object, f"{100 * observed_value:.1f}%"]

    @override
    def _validate(
        self,
        metrics: dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        return self._validate_metric_value_between(
            metric_name="column.non_null_proportion",
            metrics=metrics,
            runtime_configuration=runtime_configuration,
            execution_engine=execution_engine,
        )



from __future__ import annotations

from great_expectations.exceptions import GreatExpectationsTypeError


class InvalidRenderedContentError(GreatExpectationsTypeError):
    pass


class InlineRendererError(GreatExpectationsTypeError):
    pass


class RendererConfigurationError(GreatExpectationsTypeError):
    pass



import importlib

import pytest

from great_expectations.util import verify_dynamic_loading_support


@pytest.mark.unit
def test_store_dynamic_loading_enablement() -> None:
    module = importlib.import_module("great_expectations.data_context.store")
    module_dict = vars(module)
    for module_name in module_dict:
        if module_name.endswith("_store") or module_name.endswith("_store_backend"):
            verify_dynamic_loading_support(
                module_name=f".{module_name}",
                package_name="great_expectations.data_context.store",
            )



from __future__ import annotations

import difflib
import logging
import pathlib
import random
from pprint import pformat as pf
from typing import TYPE_CHECKING

import pytest

import great_expectations.expectations as gxe
from great_expectations import get_context
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.partitioners import ColumnPartitionerMonthly
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context import CloudDataContext, FileDataContext
from great_expectations.datasource.fluent import SqliteDatasource
from great_expectations.datasource.fluent.config import GxConfig
from great_expectations.datasource.fluent.interfaces import Datasource

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


YAML = YAMLHandler()

logger = logging.getLogger(__file__)


@pytest.mark.filesystem
def test_load_an_existing_config(
    cloud_storage_get_client_doubles,
    fluent_yaml_config_file: pathlib.Path,
    fluent_only_config: GxConfig,
):
    context = get_context(context_root_dir=fluent_yaml_config_file.parent, cloud_mode=False)

    assert context.fluent_config == fluent_only_config


@pytest.mark.filesystem
def test_serialize_fluent_config(
    cloud_storage_get_client_doubles,
    seeded_file_context: FileDataContext,
):
    dumped_yaml: str = seeded_file_context.fluent_config.yaml()
    print(f"  Dumped Config\n\n{dumped_yaml}\n")

    assert seeded_file_context.fluent_config.datasources

    for (
        ds_name,
        datasource,
    ) in seeded_file_context.fluent_config.get_datasources_as_dict().items():
        assert ds_name in dumped_yaml

        for asset_name in datasource.get_asset_names():
            assert asset_name in dumped_yaml


@pytest.mark.filesystem
def test_fluent_simple_validate_workflow(seeded_file_context: FileDataContext):
    datasource = seeded_file_context.data_sources.get("sqlite_taxi")
    assert isinstance(datasource, Datasource)
    partitioner = ColumnPartitionerMonthly(column_name="pickup_datetime")
    batch_request = datasource.get_asset("my_asset").build_batch_request(
        options={"year": 2019, "month": 1}, partitioner=partitioner
    )

    validator = seeded_file_context.get_validator(batch_request=batch_request)
    result = validator.expect_column_max_to_be_between(
        column="passenger_count", min_value=1, max_value=12
    )
    print(f"  results ->\n{pf(result)}")
    assert result["success"] is True


@pytest.mark.filesystem
def test_save_project_does_not_break(seeded_file_context: FileDataContext):
    print(seeded_file_context.fluent_config)
    seeded_file_context._save_project_config()


@pytest.mark.filesystem
def test_variables_save_does_not_break(seeded_file_context: FileDataContext):
    print(f"\tcontext.fluent_config ->\n{seeded_file_context.fluent_config}\n")
    print(f"\tcontext.variables ->\n{seeded_file_context.variables}")
    seeded_file_context.variables.save()


@pytest.mark.filesystem
def test_save_datacontext_persists_fluent_config(
    file_dc_config_dir_init: pathlib.Path, fluent_only_config: GxConfig
):
    config_file = file_dc_config_dir_init / FileDataContext.GX_YML

    initial_yaml = config_file.read_text()
    for ds_name in fluent_only_config.get_datasource_names():
        assert ds_name not in initial_yaml

    context: FileDataContext = get_context(context_root_dir=config_file.parent, cloud_mode=False)

    context.fluent_config = fluent_only_config
    context._save_project_config()

    final_yaml = config_file.read_text()
    diff = difflib.ndiff(initial_yaml.splitlines(), final_yaml.splitlines())

    print("\n".join(diff))

    for ds_name in fluent_only_config.get_datasource_names():
        assert ds_name in final_yaml


@pytest.mark.filesystem
def test_file_context_add_and_save_fluent_datasource(
    file_dc_config_dir_init: pathlib.Path,
    fluent_only_config: GxConfig,
    sqlite_database_path: pathlib.Path,
):
    datasource_name = "save_ds_test"
    config_file = file_dc_config_dir_init / FileDataContext.GX_YML

    initial_yaml = config_file.read_text()
    assert datasource_name not in initial_yaml

    context: FileDataContext = get_context(context_root_dir=config_file.parent, cloud_mode=False)

    ds = context.data_sources.add_sqlite(
        name=datasource_name, connection_string=f"sqlite:///{sqlite_database_path}"
    )

    final_yaml = config_file.read_text()
    diff = difflib.ndiff(initial_yaml.splitlines(), final_yaml.splitlines())

    print("\n".join(diff))

    assert datasource_name == ds.name
    assert datasource_name in final_yaml
    # ensure comments preserved
    assert "# Welcome to Great Expectations!" in final_yaml


# Test markers come from empty_contexts fixture
def test_context_add_and_save_fluent_datasource(
    empty_contexts: CloudDataContext | FileDataContext,
    sqlite_database_path: pathlib.Path,
):
    context = empty_contexts

    datasource_name = "save_ds_test"

    context.data_sources.add_sqlite(
        name=datasource_name, connection_string=f"sqlite:///{sqlite_database_path}"
    )

    assert datasource_name in context.data_sources.all()


# Test markers come from empty_contexts fixture
def test_context_add_or_update_datasource(
    empty_contexts: CloudDataContext | FileDataContext,
    sqlite_database_path: pathlib.Path,
):
    context = empty_contexts

    datasource: SqliteDatasource = context.data_sources.add_sqlite(
        name="save_ds_test", connection_string=f"sqlite:///{sqlite_database_path}"
    )

    assert datasource.connection_string == f"sqlite:///{sqlite_database_path}"

    # modify the datasource
    datasource.connection_string = "sqlite:///"  # type: ignore[assignment] # FIXME CoP
    context.data_sources.add_or_update_sqlite(datasource)

    updated_datasource = context.data_sources.all()[datasource.name]
    assert isinstance(updated_datasource, SqliteDatasource)
    assert updated_datasource.connection_string == "sqlite:///"


@pytest.fixture
def random_datasource(seeded_file_context: FileDataContext) -> Datasource:
    datasource = random.choice(list(seeded_file_context.fluent_datasources.values()))
    logger.info(f"Random DS - {pf(datasource.dict(), depth=1)}")
    return datasource


@pytest.mark.filesystem
def test_sources_delete_removes_datasource_from_yaml(
    random_datasource: Datasource,
    seeded_file_context: FileDataContext,
):
    print(f"Delete -> '{random_datasource.name}'\n")

    seeded_file_context.data_sources.delete(random_datasource.name)

    yaml_path = pathlib.Path(
        seeded_file_context.root_directory, seeded_file_context.GX_YML
    ).resolve(strict=True)
    yaml_contents = YAML.load(yaml_path.read_text())
    print(f"{pf(yaml_contents, depth=2)}")

    assert random_datasource.name not in yaml_contents["fluent_datasources"]  # type: ignore[operator] # always dict


@pytest.mark.filesystem
def test_ctx_delete_removes_datasource_from_yaml(
    random_datasource: Datasource, seeded_file_context: FileDataContext
):
    print(f"Delete -> '{random_datasource.name}'\n")

    seeded_file_context.delete_datasource(random_datasource.name)

    yaml_path = pathlib.Path(
        seeded_file_context.root_directory, seeded_file_context.GX_YML
    ).resolve(strict=True)
    yaml_contents = YAML.load(yaml_path.read_text())
    print(f"{pf(yaml_contents, depth=2)}")

    assert random_datasource.name not in yaml_contents["fluent_datasources"]  # type: ignore[operator] # always dict


# Test markers come from empty_contexts fixture
def test_quickstart_workflow(
    empty_contexts: CloudDataContext | FileDataContext,
    csv_path: pathlib.Path,
    mocker: MockerFixture,
):
    """
    What does this test do and why?

    Tests the Quickstart workflow noted in our docs: https://docs.greatexpectations.io/docs/tutorials/quickstart/

    In particular, this test covers the file-backend and cloud-backed usecases with this script.
    The ephemeral usecase is covered in: tests/integration/docusaurus/tutorials/quickstart/quickstart.py
    """  # noqa: E501 # FIXME CoP
    # Slight deviation from the Quickstart here:
    #   1. Using existing contexts instead of `get_context`
    #   2. Using `read_csv` on a local file instead of making a network request
    #
    # These changes should be functionally equivalent to the real workflow but be better for testing
    context = empty_contexts
    if isinstance(context, CloudDataContext):
        pytest.xfail("Checkpoint run fails in some cases on GE Cloud")

    filepath = csv_path / "yellow_tripdata_sample_2019-01.csv"
    assert filepath.exists()

    batch = context.data_sources.pandas_default.read_csv(filepath)

    # Create Expectations
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    suite.add_expectation(gxe.ExpectColumnValuesToNotBeNull(column="pickup_datetime"))
    suite.add_expectation(
        gxe.ExpectColumnValuesToBeBetween(column="passenger_count", min_value=1, max_value=6)
    )

    # Validate data
    result = batch.validate(suite)

    assert result.success

    # TODO: Add mechanism to view results.
    #       See: https://greatexpectations.atlassian.net/browse/V1-119
    # Previously we did:
    #     mock_open = mocker.patch("webbrowser.open")
    #     context.view_validation_result(result)
    #     mock_open.assert_called_once()
    # Our goal is to have:
    #     1. Setup mocks if necessary
    #     2. result.open_docs()
    #     3. assert docs opened.


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



import pytest

from great_expectations.profile.base import Profiler


@pytest.mark.unit
def test_base_class_not_instantiable_due_to_abstract_methods():
    with pytest.raises(TypeError):
        Profiler()



from typing import Mapping, Optional

import pandas as pd
import pytest

from great_expectations.compatibility.pydantic import BaseSettings
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.sql_datasource import TableAsset
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)
from tests.integration.test_utils.data_source_config.sql import (
    SQLBatchTestSetup,
)


class SnowflakeDatasourceTestConfig(DataSourceTestConfig):
    @property
    @override
    def label(self) -> str:
        return "snowflake"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.snowflake

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        return SnowflakeBatchTestSetup(
            data=data,
            config=self,
            extra_data=extra_data,
            table_name=self.table_name,
            context=context,
            engine_manager=engine_manager,
        )


class SnowflakeConnectionConfig(BaseSettings):
    """This class retrieves these values from the environment.
    If you're testing locally, you can use your Snowflake creds
    and test against your own Snowflake account.
    """

    SNOWFLAKE_USER: str
    SNOWFLAKE_PW: str
    SNOWFLAKE_ACCOUNT: str
    SNOWFLAKE_DATABASE: str
    SNOWFLAKE_WAREHOUSE: str
    SNOWFLAKE_ROLE: str

    @property
    def connection_string(self) -> str:
        # Note: we don't specify the schema here because it will be created dynamically, and we pass
        # it into the `data_sources.add_snowflake` call.
        return (
            f"snowflake://{self.SNOWFLAKE_USER}:{self.SNOWFLAKE_PW}"
            f"@{self.SNOWFLAKE_ACCOUNT}/{self.SNOWFLAKE_DATABASE}"
            f"?warehouse={self.SNOWFLAKE_WAREHOUSE}&role={self.SNOWFLAKE_ROLE}"
        )


class SnowflakeBatchTestSetup(SQLBatchTestSetup[SnowflakeDatasourceTestConfig]):
    @property
    @override
    def connection_string(self) -> str:
        return self.snowflake_connection_config.connection_string

    @property
    @override
    def use_schema(self) -> bool:
        return True

    def __init__(
        self,
        config: SnowflakeDatasourceTestConfig,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        table_name: Optional[str] = None,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> None:
        self.snowflake_connection_config = SnowflakeConnectionConfig()  # type: ignore[call-arg]  # retrieves env vars
        super().__init__(
            config=config,
            data=data,
            extra_data=extra_data,
            table_name=table_name,
            engine_manager=engine_manager,
            context=context,
        )

    @override
    def make_asset(self) -> TableAsset:
        schema = self.schema
        assert schema
        return self.context.data_sources.add_snowflake(
            name=self._random_resource_name(),
            account=self.snowflake_connection_config.SNOWFLAKE_ACCOUNT,
            user=self.snowflake_connection_config.SNOWFLAKE_USER,
            password=self.snowflake_connection_config.SNOWFLAKE_PW,
            database=self.snowflake_connection_config.SNOWFLAKE_DATABASE,
            schema=schema,
            warehouse=self.snowflake_connection_config.SNOWFLAKE_WAREHOUSE,
            role=self.snowflake_connection_config.SNOWFLAKE_ROLE,
        ).add_table_asset(
            name=self._random_resource_name(),
            table_name=self.table_name,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from great_expectations.core.batch import BatchData

if TYPE_CHECKING:
    import pandas as pd


class PandasBatchData(BatchData):
    def __init__(self, execution_engine, dataframe: pd.DataFrame) -> None:
        super().__init__(execution_engine=execution_engine)
        self._dataframe = dataframe

    @property
    def dataframe(self):
        return self._dataframe



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, List, Type

from great_expectations.core.id_dict import BatchSpec

if TYPE_CHECKING:
    from great_expectations.core.batch import LegacyBatchDefinition
    from great_expectations.datasource.fluent import BatchRequest


logger = logging.getLogger(__name__)


# noinspection SpellCheckingInspection
class DataConnector(ABC):
    """The abstract base class for all Data Connectors.

    Data Connectors produce identifying information, called Batch Specs, that Execution Engines
    can use to get individual batches of data. They add flexibility in how to obtain data
    such as with time-based partitioning, downsampling, or other techniques appropriate
    for the Datasource.

    For example, a DataConnector could produce a SQL query that logically represents "rows in
    the Events table with a timestamp on February 7, 2012," which an SqlAlchemy Datasource
    could use to materialize a SqlAlchemy Dataset corresponding to that Batch of data and
    ready for validation.

    A Batch is a sample from a data asset, sliced according to a particular rule. For example,
    an hourly slide of the Events table or “most recent Users records.” It is the primary
    unit of validation in the Great Expectations Data Context. Batches include metadata that
    identifies how they were constructed--the same Batch Spec assembled by the data connector.
    While not every Datasource will enable re-fetching a specific batch of data, GX can store
    snapshots of batches or store metadata from an external data version control system.

    Args:
        datasource_name: The name of the Datasource associated with this DataConnector instance
        data_asset_name: The name of the DataAsset using this DataConnector instance
    """

    # needed to select the asset level kwargs needed to build the DataConnector
    asset_level_option_keys: ClassVar[tuple[str, ...]] = ()
    asset_options_type: ClassVar[Type] = dict

    def __init__(
        self,
        datasource_name: str,
        data_asset_name: str,
    ) -> None:
        self._datasource_name: str = datasource_name
        self._data_asset_name: str = data_asset_name

    @property
    def data_asset_name(self) -> str:
        return self._data_asset_name

    @property
    def datasource_name(self) -> str:
        return self._datasource_name

    @abstractmethod
    def get_batch_definition_list(self, batch_request: BatchRequest) -> List[LegacyBatchDefinition]:
        """
        This interface method, implemented by subclasses, examines "BatchRequest" and converts it to one or more
        "BatchDefinition" objects, each of which can be later converted to ExecutionEngine-specific "BatchSpec" object
        for loading "Batch" of data.

        Args:
            batch_request: (BatchRequest) input "BatchRequest" object

        Returns:
            List[BatchDefinition] -- list of "BatchDefinition" objects, each corresponding to "Batch" of data downstream
        """  # noqa: E501 # FIXME CoP
        pass

    def build_batch_spec(self, batch_definition: LegacyBatchDefinition) -> BatchSpec:
        """
        Builds batch_spec from batch_definition by generating batch_spec params and adding any pass_through params

        Args:
            batch_definition (LegacyBatchDefinition): required batch_definition parameter for retrieval
        Returns:
            BatchSpec object built from BatchDefinition
        """  # noqa: E501 # FIXME CoP
        batch_spec_params: dict = self._generate_batch_spec_parameters_from_batch_definition(
            batch_definition=batch_definition
        )
        batch_spec = BatchSpec(**batch_spec_params)
        return batch_spec

    def test_connection(self) -> bool:
        """Test the connection to data, accessible to the present "DataConnector" object.

        Raises:
            bool: True of connection test succeeds; False, otherwise.
        """
        return self.get_unmatched_data_reference_count() < self.get_data_reference_count()

    @abstractmethod
    def get_data_references(self) -> List[Any]:
        """
        This interface method lists objects in the underlying data store used to create a list of data_references (type depends on cloud storage environment, SQL DBMS, etc.).
        """  # noqa: E501 # FIXME CoP
        pass

    @abstractmethod
    def get_data_reference_count(self) -> int:
        """
        This interface method returns number of all (e.g., cached) data references (useful for diagnostics).

        Returns:
            int -- number of data references identified
        """  # noqa: E501 # FIXME CoP
        pass

    @abstractmethod
    def get_matched_data_references(self) -> List[Any]:
        """
        This interface method returns (e.g., cached) data references that were successfully matched based on "BatchRequest" options.

        Returns:
            List[Any] -- unmatched data references (type depends on cloud storage environment, SQL DBMS, etc.)
        """  # noqa: E501 # FIXME CoP
        pass

    @abstractmethod
    def get_matched_data_reference_count(self) -> int:
        """
        This interface method returns number of all (e.g., cached) matched data references (useful for diagnostics).

        Returns:
            int -- number of data references identified
        """  # noqa: E501 # FIXME CoP
        pass

    @abstractmethod
    def get_unmatched_data_references(self) -> List[Any]:
        """
        This interface method returns (e.g., cached) data references that could not be matched based on "BatchRequest" options.

        Returns:
            List[Any] -- unmatched data references (type depends on cloud storage environment, SQL DBMS, etc.)
        """  # noqa: E501 # FIXME CoP
        pass

    @abstractmethod
    def get_unmatched_data_reference_count(self) -> int:
        """
        This interface method returns number of all (e.g., cached) unmatched data references (useful for diagnostics).

        Returns:
            int -- number of data references identified
        """  # noqa: E501 # FIXME CoP
        pass

    @abstractmethod
    def _generate_batch_spec_parameters_from_batch_definition(
        self, batch_definition: LegacyBatchDefinition
    ) -> dict:
        """
        This interface method, implemented by subclasses, examines "BatchDefinition" and converts it to
        ExecutionEngine-specific "BatchSpec" object for loading "Batch" of data.  Implementers will typically define
        their own interfaces that their subclasses must implement in order to provide storage-specific specifics.

        Args:
            batch_definition: (BatchDefinition) input "BatchRequest" object

        Returns:
            dict -- dictionary of "BatchSpec" properties
        """  # noqa: E501 # FIXME CoP
        pass

    @staticmethod
    def _batch_definition_matches_batch_request(
        batch_definition: LegacyBatchDefinition, batch_request: BatchRequest
    ) -> bool:
        if not (
            batch_request.datasource_name == batch_definition.datasource_name
            and batch_request.data_asset_name == batch_definition.data_asset_name
        ):
            return False

        if batch_request.options:
            for key, value in batch_request.options.items():
                if value is not None and not (
                    (key in batch_definition.batch_identifiers)
                    and (batch_definition.batch_identifiers[key] == batch_request.options[key])
                ):
                    return False

        return True



import pandas

from great_expectations.metrics.column.mean import (
    ColumnMean,
    ColumnMeanResult,
)
from great_expectations.metrics.metric_results import MetricErrorResult
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

DATA_FRAME = pandas.DataFrame(
    {
        "id": [1, 2, 3, 4],
        "number": [1, 2, 3, 4],
        "string": ["a", "b", "c", "d"],
    },
)

DATA_SOURCES_WITHOUT_SPARK_DATABRICKS_SQLITE: list[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    PandasDataFrameDatasourceTestConfig(),
]

DATA_SOURCES: list[DataSourceTestConfig] = DATA_SOURCES_WITHOUT_SPARK_DATABRICKS_SQLITE + [
    SparkFilesystemCsvDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]


@parameterize_batch_for_data_sources(
    data_source_configs=DATA_SOURCES,
    data=DATA_FRAME,
)
def test_mean_success(batch_for_datasource) -> None:
    batch = batch_for_datasource
    metric = ColumnMean(column="number")
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, ColumnMeanResult)
    assert metric_result.value == 2.5


# For spark, when computing the mean, if it fails, the metric name changes from
# `column.mean` to `column.aggregate.mean`.
# There is a bug to track fixing this: https://greatexpectations.atlassian.net/browse/GX-448
# For databricks, when computing the mean, any non-numeric values are ignored and the result is
# None, which will cause a crash later when trying to set the value of the MetricResult
# (not MetricErrorResult) to None.
# For sqlite, when computing the mean, any non-numeric values are ignored (or maybe treated
# as 0) so we don't an error.
@parameterize_batch_for_data_sources(
    data_source_configs=DATA_SOURCES_WITHOUT_SPARK_DATABRICKS_SQLITE,
    data=DATA_FRAME,
)
def test_mean_failure(batch_for_datasource) -> None:
    batch = batch_for_datasource
    metric = ColumnMean(column="string")
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, MetricErrorResult)
